# -*- coding: utf-8 -*-

from collections.abc import Generator, Iterator
from typing import (
	TYPE_CHECKING,
)

if TYPE_CHECKING:
	import sqlite3

	from pyglossary.glossary_types import EntryType, GlossaryType
	from pyglossary.xdxf.transform import XdxfTransformer

from pyglossary.core import log
from pyglossary.option import BoolOption, Option

__all__ = [
	"Reader",
	"Writer",
	"description",
	"enable",
	"extensionCreate",
	"extensions",
	"format",
	"kind",
	"lname",
	"optionsProp",
	"singleFile",
	"website",
	"wiki",
]

enable = True
lname = "ayandict_sqlite"
format = "AyanDictSQLite"
description = "AyanDict SQLite"
extensions = ()
extensionCreate = ".db"
singleFile = True
kind = "binary"
wiki = ""
website = (
	"https://github.com/ilius/ayandict",
	"ilius/ayandict",
)
optionsProp: "dict[str, Option]" = {
	"fuzzy": BoolOption(
		comment="Create fuzzy search data",
	),
}


class Reader:
	def __init__(self, glos: "GlossaryType") -> None:
		self._glos = glos
		self._clear()

	def _clear(self) -> None:
		self._filename = ""
		self._con: "sqlite3.Connection | None" = None
		self._cur: "sqlite3.Cursor | None" = None

	def open(self, filename: str) -> None:
		from sqlite3 import connect

		self._filename = filename
		self._con = connect(filename)
		self._cur = self._con.cursor()
		self._glos.setDefaultDefiFormat("h")

		self._cur.execute("SELECT key, value FROM meta;")
		for row in self._cur.fetchall():
			if row[0] == "hash":
				continue
			self._glos.setInfo(row[0], row[1])

	def __len__(self) -> int:
		if self._cur is None:
			raise ValueError("cur is None")
		self._cur.execute("select count(id) from entry")
		return self._cur.fetchone()[0]

	def __iter__(self) -> "Iterator[EntryType]":
		from json import loads

		if self._cur is None:
			raise ValueError("cur is None")
		self._cur.execute(
			"SELECT entry.term, entry.article, "
			"json_group_array(alt.term)"
			"FROM entry LEFT JOIN alt ON entry.id=alt.id "
			"GROUP BY entry.id;",
		)
		for row in self._cur.fetchall():
			terms = [row[0]] + [alt for alt in loads(row[2]) if alt]
			article = row[1]
			yield self._glos.newEntry(terms, article, defiFormat="h")

	def close(self) -> None:
		if self._cur:
			self._cur.close()
		if self._con:
			self._con.close()
		self._clear()


class Writer:
	_fuzzy: int = True

	def __init__(self, glos: "GlossaryType") -> None:
		self._glos = glos
		self._clear()

	def _clear(self) -> None:
		self._filename = ""
		self._con: "sqlite3.Connection | None" = None
		self._cur: "sqlite3.Cursor | None" = None
		self._xdxfTr: "XdxfTransformer | None" = None

	def open(self, filename: str) -> None:
		from sqlite3 import connect

		self._filename = filename
		con = self._con = connect(filename)
		self._cur = self._con.cursor()

		for query in (
			"CREATE TABLE meta ('key' TEXT PRIMARY KEY NOT NULL, 'value' TEXT);",
			(
				"CREATE TABLE entry ('id' INTEGER PRIMARY KEY NOT NULL, "
				"'term' TEXT, 'article' TEXT);"
			),
			"CREATE TABLE alt ('id' INTEGER NOT NULL, 'term' TEXT);",
			"CREATE INDEX idx_meta ON meta(key);",
			"CREATE INDEX idx_entry_term ON entry(term COLLATE NOCASE);",
			"CREATE INDEX idx_alt_id ON alt(id);",
			"CREATE INDEX idx_alt_term ON alt(term COLLATE NOCASE);",
		):
			try:
				con.execute(query)
			except Exception as e:  # noqa: PERF203
				log.error(f"query: {query}")
				raise e

		for key, value in self._glos.iterInfo():
			con.execute(
				"INSERT	INTO meta (key, value) VALUES (?, ?);",
				(key, value),
			)

		if self._fuzzy:
			con.execute(
				"CREATE TABLE fuzzy3 ('sub' TEXT NOT NULL, "
				"'term' TEXT NOT NULL, "
				"id INTEGER NOT NULL);",
			)
			con.execute(
				"CREATE INDEX idx_fuzzy3_sub ON fuzzy3(sub COLLATE NOCASE);",
			)

		con.commit()

	def finish(self) -> None:
		if self._con is None or self._cur is None:
			return

		self._con.commit()
		self._con.close()
		self._con = None
		self._cur = None

	def xdxf_setup(self) -> None:
		from pyglossary.xdxf.transform import XdxfTransformer

		# if self._xsl:
		# 	self._xdxfTr = XslXdxfTransformer(encoding="utf-8")
		# 	return
		self._xdxfTr = XdxfTransformer(encoding="utf-8")

	def xdxf_transform(self, text: str) -> str:
		if self._xdxfTr is None:
			self.xdxf_setup()
		return self._xdxfTr.transformByInnerString(text)  # type: ignore

	def write(self) -> "Generator[None, EntryType, None]":
		import hashlib

		cur = self._cur
		if cur is None:
			raise ValueError("cur is None")
		_hash = hashlib.md5()
		while True:
			entry = yield
			if entry is None:
				break
			if entry.isData():
				# can save it with entry.save(directory)
				continue
			defi = entry.defi
			entry.detectDefiFormat()
			if entry.defiFormat == "m":
				if "\n" in defi:
					defi = f"<pre>{defi}</pre>"
			elif entry.defiFormat == "x":
				defi = self.xdxf_transform(defi)

			cur.execute(
				"INSERT INTO entry(term, article) VALUES (?, ?);",
				(entry.l_word[0], defi),
			)
			_id = cur.lastrowid
			if _id is None:
				raise ValueError("lastrowid is None")
			for alt in entry.l_word[1:]:
				cur.execute(
					"INSERT INTO alt(id, term) VALUES (?, ?);",
					(_id, alt),
				)
			_hash.update(entry.s_word.encode("utf-8"))
			if self._fuzzy:
				self.addFuzzy(_id, entry.l_word)

		cur.execute(
			"INSERT INTO meta (key, value) VALUES (?, ?);",
			("hash", _hash.hexdigest()),
		)

	def addFuzzy(self, _id: int, terms: list[str]) -> None:
		cur = self._cur
		if cur is None:
			raise ValueError("cur is None")
		for term in terms:
			subs: "set[str]" = set()
			for word in term.split(" "):
				eword = "\n" + word
				subs.update(eword[i : i + 3] for i in range(len(eword) - 2))
			for sub in subs:
				cur.execute(
					"INSERT INTO fuzzy3(sub, term, id) VALUES (?, ?, ?);",
					(sub, term, _id),
				)



# -*- coding: utf-8 -*-

import html
from collections.abc import Callable, Iterator
from operator import itemgetter
from typing import TYPE_CHECKING, cast

if TYPE_CHECKING:
	import sqlite3

	from pyglossary.glossary_types import EntryType, GlossaryType
	from pyglossary.lxml_types import Element, T_htmlfile
	from pyglossary.option import Option


from pyglossary.core import log

__all__ = [
	"Reader",
	"description",
	"enable",
	"extensionCreate",
	"extensions",
	"format",
	"kind",
	"lname",
	"optionsProp",
	"singleFile",
	"website",
	"wiki",
]

enable = True
lname = "dict_cc"
format = "Dictcc"
description = "Dict.cc (SQLite3)"
extensions = ()
extensionCreate = ".db"
singleFile = True
kind = "binary"
wiki = "https://en.wikipedia.org/wiki/Dict.cc"
website = (
	"https://play.google.com/store/apps/details?id=cc.dict.dictcc",
	"dict.cc dictionary - Google Play",
)
optionsProp: "dict[str, Option]" = {}


class Reader:
	def __init__(self, glos: "GlossaryType") -> None:
		self._glos = glos
		self._clear()

	def _clear(self) -> None:
		self._filename = ""
		self._con: "sqlite3.Connection | None" = None
		self._cur: "sqlite3.Cursor | None" = None

	def open(self, filename: str) -> None:
		from sqlite3 import connect

		self._filename = filename
		self._con = connect(filename)
		self._cur = self._con.cursor()
		self._glos.setDefaultDefiFormat("h")

	def __len__(self) -> int:
		if self._cur is None:
			raise ValueError("cur is None")
		self._cur.execute(
			"select count(distinct term1)+count(distinct term2) from main_ft",
		)
		return self._cur.fetchone()[0]

	@staticmethod
	def makeList(
		hf: "T_htmlfile",
		input_elements: "list[Element]",
		processor: "Callable",
		single_prefix: str = "",
		skip_single: bool = True,
	) -> None:
		"""Wrap elements into <ol> if more than one element."""
		if not input_elements:
			return

		if skip_single and len(input_elements) == 1:
			hf.write(single_prefix)
			processor(hf, input_elements[0])
			return

		with hf.element("ol"):
			for el in input_elements:
				with hf.element("li"):
					processor(hf, el)

	@staticmethod
	def makeGroupsList(
		hf: "T_htmlfile",
		groups: "list[tuple[str, str]]",
		processor: "Callable[[T_htmlfile, tuple[str, str]], None]",
		single_prefix: str = "",
		skip_single: bool = True,
	) -> None:
		"""Wrap elements into <ol> if more than one element."""
		if not groups:
			return

		if skip_single and len(groups) == 1:
			hf.write(single_prefix)
			processor(hf, groups[0])
			return

		with hf.element("ol"):
			for el in groups:
				with hf.element("li"):
					processor(hf, el)

	def writeSense(  # noqa: PLR6301
		self,
		hf: "T_htmlfile",
		row: "tuple[str, str]",
	) -> None:
		from lxml import etree as ET

		trans, entry_type = row
		if entry_type:
			with hf.element("i"):
				hf.write(f"{entry_type}")
			hf.write(ET.Element("br"))
		try:
			hf.write(trans + " ")
		except Exception as e:
			log.error(f"error in writing {trans!r}, {e}")
			hf.write(repr(trans) + " ")
		else:
			with hf.element("big"):
				with hf.element("a", href=f"bword://{trans}"):
					hf.write("⏎")

	def iterRows(
		self,
		column1: str,
		column2: str,
	) -> "Iterator[tuple[str, str, str]]":
		if self._cur is None:
			raise ValueError("cur is None")
		self._cur.execute(
			f"select {column1}, {column2}, entry_type from main_ft"
			f" order by {column1}",
		)
		for row in self._cur.fetchall():
			term1 = row[0]
			term2 = row[1]
			try:
				term1 = html.unescape(term1)
			except Exception as e:
				log.error(f"html.unescape({term1!r}) -> {e}")
			try:
				term2 = html.unescape(term2)
			except Exception as e:
				log.error(f"html.unescape({term2!r}) -> {e}")
			yield term1, term2, row[2]

	def parseGender(self, headword: str) -> "tuple[str | None, str]":  # noqa: PLR6301
		# {m}	masc	masculine	German: maskulin
		# {f}	fem 	feminine	German: feminin
		# {n}	neut	neutral		German: neutral
		# { }	????
		i = headword.find(" {")
		if i <= 0:
			return None, headword
		if len(headword) < i + 4:
			return None, headword
		if headword[i + 3] != "}":
			return None, headword
		g = headword[i + 2]
		gender = None
		if g == "m":
			gender = "masculine"
		elif g == "f":
			gender = "feminine"
		elif g == "n":
			gender = "neutral"
		else:
			log.warning(f"invalid gender {g!r}")
			return None, headword
		headword = headword[:i] + headword[i + 4 :]
		return gender, headword

	def _iterOneDirection(
		self,
		column1: str,
		column2: str,
	) -> "Iterator[EntryType]":
		from io import BytesIO
		from itertools import groupby

		from lxml import etree as ET

		glos = self._glos
		for headwordEscaped, groupsOrig in groupby(
			self.iterRows(column1, column2),
			key=itemgetter(0),
		):
			headword = html.unescape(headwordEscaped)
			groups: "list[tuple[str, str]]" = [
				(term2, entry_type) for _, term2, entry_type in groupsOrig
			]
			f = BytesIO()
			gender, headword = self.parseGender(headword)
			with ET.htmlfile(f, encoding="utf-8") as hf:
				with hf.element("div"):
					if gender:
						with hf.element("i"):
							hf.write(gender)
						hf.write(ET.Element("br"))
					self.makeGroupsList(
						cast("T_htmlfile", hf),
						groups,
						self.writeSense,
					)
			defi = f.getvalue().decode("utf-8")
			yield glos.newEntry(headword, defi, defiFormat="h")

	def __iter__(self) -> "Iterator[EntryType]":
		yield from self._iterOneDirection("term1", "term2")
		yield from self._iterOneDirection("term2", "term1")

	def close(self) -> None:
		if self._cur:
			self._cur.close()
		if self._con:
			self._con.close()
		self._clear()



import json
from collections import OrderedDict
from typing import TYPE_CHECKING

if TYPE_CHECKING:
	from typing import AnyStr, TypeAlias

__all__ = ["dataToPrettyJson", "jsonToData", "jsonToOrderedData"]

JsonEncodable: "TypeAlias" = "dict | list"
# OrderedDict is also subclass of Dict, issubclass(OrderedDict, Dict) is True


def dataToPrettyJson(
	data: JsonEncodable,
	ensure_ascii: bool = False,
	sort_keys: bool = False,
) -> str:
	return json.dumps(
		data,
		sort_keys=sort_keys,
		indent="\t",
		ensure_ascii=ensure_ascii,
	)


def jsonToData(st: "AnyStr") -> JsonEncodable:
	return json.loads(st)


def jsonToOrderedData(text: str) -> "OrderedDict":
	return json.JSONDecoder(
		object_pairs_hook=OrderedDict,
	).decode(text)



# -*- coding: utf-8 -*-

from collections.abc import Iterator
from typing import TYPE_CHECKING

from pyglossary.glossary_types import (
	EntryType,
	GlossaryType,
)
from pyglossary.info_writer import InfoWriter as Writer

if TYPE_CHECKING:
	from pyglossary.option import Option

__all__ = [
	"Reader",
	"Writer",
	"description",
	"enable",
	"extensionCreate",
	"extensions",
	"format",
	"kind",
	"lname",
	"optionsProp",
	"singleFile",
	"website",
	"wiki",
]

enable = True
lname = "info"
format = "Info"
description = "Glossary Info (.info)"
extensions = (".info",)
extensionCreate = ".info"
singleFile = True
kind = "text"
wiki = ""
website = None

# key is option/argument name, value is instance of Option
optionsProp: "dict[str, Option]" = {}


class Reader:
	def __init__(self, glos: GlossaryType) -> None:
		self._glos = glos

	def close(self) -> None:
		pass

	def open(self, filename: str) -> None:
		from pyglossary.json_utils import jsonToOrderedData

		with open(filename, encoding="utf-8") as infoFp:
			info = jsonToOrderedData(infoFp.read())
		for key, value in info.items():
			self._glos.setInfo(key, value)

	def __len__(self) -> int:
		return 0

	def __iter__(self) -> "Iterator[EntryType | None]":
		yield None



# -*- coding: utf-8 -*-

import json
import re
from collections.abc import Generator, Sequence
from typing import Any

from pyglossary import os_utils
from pyglossary.flags import ALWAYS
from pyglossary.glossary_types import EntryType, GlossaryType
from pyglossary.option import (
	BoolOption,
	IntOption,
	Option,
	StrOption,
)

__all__ = [
	"Writer",
	"description",
	"enable",
	"extensionCreate",
	"extensions",
	"format",
	"kind",
	"lname",
	"optionsProp",
	"singleFile",
	"website",
	"wiki",
]

enable = True
lname = "yomichan"
format = "Yomichan"
description = "Yomichan (.zip)"
extensions = (".zip",)
extensionCreate = ".zip"
singleFile = True
sortOnWrite = ALWAYS
sortKeyName = "headword"
kind = "package"
wiki = ""
website = (
	"https://foosoft.net/projects/yomichan/",
	"foosoft.net",
)
optionsProp: "dict[str, Option]" = {
	"term_bank_size": IntOption(
		comment="The number of terms in each term bank json file.",
	),
	"term_from_headword_only": BoolOption(
		comment=(
			"If set to true, only create a term for the headword for each entry, "
			"as opposed to create one term for each alternate word. "
			"If the headword is ignored by the `ignore_word_with_pattern` option, "
			"the next word in the alternate list that is not ignored is used as "
			"headword."
		),
	),
	"no_term_from_reading": BoolOption(
		comment=(
			"When there are multiple alternate words, don't create term for the "
			"one that is the same as the the reading form, which is chosen to be "
			"the first alternate forms that consists solely of Hiragana and "
			"Katakana. "
			"For example, an entry could contain both 'だいがく' and '大学' as "
			"alternate words. Setting this option to true would prevent a term "
			"to be created for the former."
		),
	),
	"delete_word_pattern": StrOption(
		comment=(
			"When given, all non-overlapping matches of this regular expression "
			"are removed from word strings. "
			"For example, if an entry has word 'あま·い', setting the "
			"pattern to `·` removes all center dots, or more precisely use "
			"`·(?=[\\u3040-\\u309F])` to only remove center dots that precede "
			"Hiragana characters. Either way, the original word is replaced "
			"with 'あまい'."
		),
	),
	"ignore_word_with_pattern": StrOption(
		comment=(
			"When given, don't create terms for a word if any of its substrings "
			"matches this regular expression. "
			"For example, an entry could contain both 'だいがく【大学】' and '大学' "
			"as alternate words. Setting this option with value `r'【.+】'` would "
			"prevent a term to be created for the former."
		),
	),
	"alternates_from_word_pattern": StrOption(
		comment=(
			"When given, the regular expression is used to find additional "
			"alternate words for the same entry from matching substrings in "
			"the original words. "
			"If there are no capturing groups in the regular expression, "
			"then all matched substrings are added to the list of alternate "
			"words. "
			"If there are capturing groups, then substrings matching the groups "
			"are added to the alternate words list instead. "
			"For example, if an entry has 'だいがく【大学】' as a word, then "
			"`\\w+(?=【)` adds 'だいがく' as an additional word, while "
			"`(\\w+)【(\\w+)】` adds both 'だいがく' and '大学'."
		),
	),
	"alternates_from_defi_pattern": StrOption(
		comment=(
			"When given, the regular expression is used to find additional "
			"alternate words for the same entry from matching substrings in "
			"the definition. `^` and `$` can be used to match start and end of "
			"lines, respectively. "
			"If there are no capturing groups in the regular expression, "
			"then all matched substrings are added to the list of alternate "
			"words. "
			"If there are capturing groups, then substrings matching the groups "
			"are added to the alternate words list instead. "
			"For example, if an entry has 'だいがく【大学】' in its definition, then "
			"`\\w+【(\\w+)】` adds '大学' as an additional word."
		),
	),
	"rule_v1_defi_pattern": StrOption(
		comment=(
			"When given, if any substring of an entry's definition matches this "
			"regular expression, then the term(s) created from entry are labeled "
			"as ichidan verb. Yomichan uses this information to match conjugated "
			"forms of words. `^` and `$` can be used to match start and end of "
			"lines, respectively. "
			"For example, setting this option to `^\\(動[上下]一\\)$` identifies "
			"entries where there's a line of '(動上一)' or '(動下一)'."
		),
	),
	"rule_v5_defi_pattern": StrOption(
		comment=(
			"When given, if any substring of an entry's definition matches this "
			"regular expression, then the term(s) created from entry are labeled "
			"as godan verb. Yomichan uses this information to match conjugated "
			"forms of words. `^` and `$` can be used to match start and end of "
			"lines, respectively. "
			"For example, setting this option to `^\\(動五\\)$` identifies "
			"entries where there's a line of '(動五)'."
		),
	),
	"rule_vs_defi_pattern": StrOption(
		comment=(
			"When given, if any substring of an entry's definition matches this "
			"regular expression, then the term(s) created from entry are labeled "
			"as suru verb. Yomichan uses this information to match conjugated "
			"forms of words. `^` and `$` can be used to match start and end of "
			"lines, respectively. "
			"For example, setting this option to `^スル$` identifies entries where "
			"there's a line of 'スル'."
		),
	),
	"rule_vk_defi_pattern": StrOption(
		comment=(
			"When given, if any substring of an entry's definition matches this "
			"regular expression, then the term(s) created from entry are labeled "
			"as kuru verb. Yomichan uses this information to match conjugated "
			"forms of words. `^` and `$` can be used to match start and end of "
			"lines, respectively. "
			"For example, setting this option to `^\\(動カ変\\)$` identifies "
			"entries where there's a line of '(動カ変)'."
		),
	),
	"rule_adji_defi_pattern": StrOption(
		comment=(
			"When given, if any substring of an entry's definition matches this "
			"regular expression, then the term(s) created from entry are labeled "
			"as i-adjective. Yomichan uses this information to match conjugated "
			"forms of words. `^` and `$` can be used to match start and end of "
			"lines, respectively. "
			"For example, setting this option to `r'^\\(形\\)$'` identify "
			"entries where there's a line of '(形)'."
		),
	),
}


def _isKana(char: str) -> bool:
	assert len(char) == 1
	val = ord(char)
	return (
		0x3040 <= val <= 0x309F  # Hiragana
		or 0x30A0 <= val <= 0x30FF  # Katakana (incl. center dot)
		or 0xFF65 <= val <= 0xFF9F  # Half-width Katakana (incl. center dot)
	)


def _isKanji(char: str) -> bool:
	assert len(char) == 1
	val = ord(char)
	return (
		0x3400 <= val <= 0x4DBF  # CJK Unified Ideographs Extension A
		or 0x4E00 <= val <= 0x9FFF  # CJK Unified Ideographs
		or 0xF900 <= val <= 0xFAFF  # CJK Compatibility Ideographs
		or 0x20000 <= val <= 0x2A6DF  # CJK Unified Ideographs Extension B
		or 0x2A700 <= val <= 0x2B73F  # CJK Unified Ideographs Extension C
		or 0x2B740 <= val <= 0x2B81F  # CJK Unified Ideographs Extension D
		or 0x2F800 <= val <= 0x2FA1F  # CJK Compatibility Ideographs Supplement
	)


def _uniqueList(lst: "Sequence") -> "list[Any]":
	seen = set()
	result = []
	for elem in lst:
		if elem not in seen:
			seen.add(elem)
			result.append(elem)

	return result


class Writer:
	depends = {
		"bs4": "beautifulsoup4",
	}

	_term_bank_size = 10_000
	_term_from_headword_only = True
	_no_term_from_reading = True
	_delete_word_pattern = ""
	_ignore_word_with_pattern = ""
	_alternates_from_word_pattern = ""
	_alternates_from_defi_pattern = ""
	_rule_v1_defi_pattern = ""
	_rule_v5_defi_pattern = ""
	_rule_vs_defi_pattern = ""
	_rule_vk_defi_pattern = ""
	_rule_adji_defi_pattern = ""

	def __init__(self, glos: "GlossaryType") -> None:
		self._glos = glos
		self._filename = ""
		# Yomichan technically supports "structured content" that renders to
		# HTML, but it doesn't seem widely used. So here we also strip HTML
		# formatting for simplicity.
		glos.removeHtmlTagsAll()

	def _getInfo(self, key: str) -> str:
		info = self._glos.getInfo(key)
		return info.replace("\n", "<br>")

	def _getAuthor(self) -> str:
		return self._glos.author.replace("\n", "<br>")

	def _getDictionaryIndex(self) -> "dict[str, Any]":
		# Schema: https://github.com/FooSoft/yomichan/
		# blob/master/ext/data/schemas/dictionary-index-schema.json
		return {
			"title": self._getInfo("title"),
			"revision": "PyGlossary export",
			"sequenced": True,
			"format": 3,
			"author": self._getAuthor(),
			"url": self._getInfo("website"),
			"description": self._getInfo("description"),
		}

	def _compileRegex(self) -> None:
		for field_name in (
			"_delete_word_pattern",
			"_ignore_word_with_pattern",
			"_alternates_from_word_pattern",
			"_alternates_from_defi_pattern",
			"_rule_v1_defi_pattern",
			"_rule_v5_defi_pattern",
			"_rule_vs_defi_pattern",
			"_rule_vk_defi_pattern",
			"_rule_adji_defi_pattern",
		):
			value = getattr(self, field_name)
			if value and isinstance(value, str):
				setattr(self, field_name, re.compile(value))

	def _getExpressionsAndReadingFromEntry(
		self,
		entry: "EntryType",
	) -> "tuple[list[str], str]":
		term_expressions = list(entry.l_word)
		if self._alternates_from_word_pattern:
			for word in entry.l_word:
				term_expressions += re.findall(
					self._alternates_from_word_pattern,
					word,
				)

		if self._alternates_from_defi_pattern:
			term_expressions += re.findall(
				self._alternates_from_defi_pattern,
				entry.defi,
				re.MULTILINE,
			)

		if self._delete_word_pattern:
			term_expressions = [
				re.sub(self._delete_word_pattern, "", expression)
				for expression in term_expressions
			]

		if self._ignore_word_with_pattern:
			term_expressions = [
				expression
				for expression in term_expressions
				if not re.search(self._ignore_word_with_pattern, expression)
			]

		term_expressions = _uniqueList(term_expressions)

		try:
			reading = next(
				expression
				for expression in entry.l_word + term_expressions
				if all(map(_isKana, expression))
			)
		except StopIteration:
			reading = ""

		if self._no_term_from_reading and len(term_expressions) > 1:
			term_expressions = [
				expression for expression in term_expressions if expression != reading
			]

		if self._term_from_headword_only:
			term_expressions = term_expressions[:1]

		return term_expressions, reading

	def _getRuleIdentifiersFromEntry(self, entry: EntryType) -> list[str]:
		return [
			r
			for p, r in (
				(self._rule_v1_defi_pattern, "v1"),
				(self._rule_v5_defi_pattern, "v5"),
				(self._rule_vs_defi_pattern, "vs"),
				(self._rule_vk_defi_pattern, "vk"),
				(self._rule_adji_defi_pattern, "adj-i"),
			)
			if p and re.search(p, entry.defi, re.MULTILINE)
		]

	def _getTermsFromEntry(
		self,
		entry: "EntryType",
		sequenceNumber: int,
	) -> "list[list[Any]]":
		termExpressions, reading = self._getExpressionsAndReadingFromEntry(entry)
		ruleIdentifiers = self._getRuleIdentifiersFromEntry(entry)

		# Schema: https://github.com/FooSoft/yomichan/
		# blob/master/ext/data/schemas/dictionary-term-bank-v3-schema.json
		return [
			[
				expression,
				# reading only added if expression contains kanji
				reading if any(map(_isKanji, expression)) else "",
				"",  # definition tags
				" ".join(ruleIdentifiers),
				0,  # score
				[entry.defi],
				sequenceNumber,
				"",  # term tags
			]
			for expression in termExpressions
		]

	def open(self, filename: str) -> None:
		self._filename = filename
		self._glos.mergeEntriesWithSameHeadwordPlaintext()

	def finish(self) -> None:
		self._filename = ""

	def write(self) -> "Generator[None, EntryType, None]":
		with os_utils.indir(self._filename, create=True):
			with open("index.json", "w", encoding="utf-8") as f:
				json.dump(self._getDictionaryIndex(), f, ensure_ascii=False)

			entryCount = 0
			termBankIndex = 0
			terms: "list[list[Any]]" = []

			def flushTerms() -> None:
				nonlocal termBankIndex
				if not terms:
					return

				with open(
					f"term_bank_{termBankIndex + 1}.json",
					mode="w",
					encoding="utf-8",
				) as _file:
					json.dump(terms, _file, ensure_ascii=False)

				terms.clear()
				termBankIndex += 1

			while True:
				entry: EntryType
				entry = yield
				if entry is None:
					break

				if entry.isData():
					continue

				terms.extend(self._getTermsFromEntry(entry, entryCount))
				entryCount += 1
				if len(terms) >= self._term_bank_size:
					flushTerms()

			flushTerms()



# -*- coding: utf-8 -*-

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License <http://www.gnu.org/licenses/gpl-3.0.txt>
# for more details.
#
# Copyright (C) 2023 Saeed Rasooli
# Copyright (C) 2015 Igor Tkach
#
# This plugin is based on https://github.com/itkach/wordnet2slob

import io
import os
import re
import sys
from collections import defaultdict
from collections.abc import Iterator
from typing import TYPE_CHECKING

from pyglossary.core import log

if TYPE_CHECKING:
	from pyglossary.glossary_types import EntryType, GlossaryType
	from pyglossary.option import Option

__all__ = [
	"Reader",
	"description",
	"enable",
	"extensionCreate",
	"extensions",
	"format",
	"kind",
	"lname",
	"optionsProp",
	"singleFile",
	"website",
	"wiki",
]

enable = True
lname = "wordnet"
format = "Wordnet"
description = "WordNet"
extensions = ()
extensionCreate = ""
singleFile = False
kind = "directory"
wiki = "https://en.wikipedia.org/wiki/WordNet"
website = (
	"https://wordnet.princeton.edu/",
	"WordNet - A Lexical Database for English",
)

# key is option/argument name, value is instance of Option
optionsProp: "dict[str, Option]" = {}

# original expression from
# http://stackoverflow.com/questions/694344/regular-expression-that-matches-between-quotes-containing-escaped-quotes
# "(?:[^\\"]+|\\.)*"
# some examples don't have closing quote which
# make the subn with this expression hang
# quotedTextPattern = re.compile(r'"(?:[^"]+|\.)*["|\n]')

# make it a capturing group so that we can get rid of quotes
quotedTextPattern = re.compile(r'"([^"]+)"')

refPattern = re.compile(r"`(\w+)'")


class SynSet:
	def __init__(self, line: str | bytes) -> None:
		self.line = line
		if isinstance(line, bytes):
			line = line.decode("utf-8")
		meta, self.gloss = line.split("|")
		self.meta_parts = meta.split()

	@property
	def offset(self) -> int:
		return int(self.meta_parts[0])

	@property
	def lex_filenum(self) -> str:
		return self.meta_parts[1]

	@property
	def ss_type(self) -> str:
		return self.meta_parts[2]

	@property
	def w_cnt(self) -> int:
		return int(self.meta_parts[3], 16)

	@property
	def words(self) -> list[str]:
		return [self.meta_parts[4 + 2 * i].replace("_", " ") for i in range(self.w_cnt)]

	@property
	def pointers(self) -> "list[Pointer]":
		p_cnt_index = 4 + 2 * self.w_cnt
		p_cnt = self.meta_parts[p_cnt_index]
		pointer_count = int(p_cnt)
		start = p_cnt_index + 1
		return [
			Pointer(*self.meta_parts[start + i * 4 : start + (i + 1) * 4])  # type: ignore
			for i in range(pointer_count)
		]

	def __repr__(self) -> str:
		return f"SynSet({self.line!r})"


class PointerSymbols:
	n = {
		"!": "Antonyms",
		"@": "Hypernyms",
		"@i": "Instance hypernyms",
		"~": "Hyponyms",
		"~i": "Instance hyponyms",
		"#m": "Member holonyms",
		"#s": "Substance holonyms",
		"#p": "Part holonyms",
		"%m": "Member meronyms",
		"%s": "Substance meronyms",
		"%p": "Part meronyms",
		"=": "Attributes",
		"+": "Derivationally related forms",
		";c": "Domain of synset - TOPIC",
		"-c": "Member of this domain - TOPIC",
		";r": "Domain of synset - REGION",
		"-r": "Member of this domain - REGION",
		";u": "Domain of synset - USAGE",
		"-u": "Member of this domain - USAGE",
		"^": "Also see",
	}

	v = {
		"!": "Antonyms",
		"@": "Hypernyms",
		"~": "Hyponyms",
		"*": "Entailments",
		">": "Cause",
		"^": "Also see",
		"$": "Verb group",
		"+": "Derivationally related forms",
		";c": "Domain of synset - TOPIC",
		";r": "Domain of synset - REGION",
		";u": "Domain of synset - USAGE",
	}

	a = s = {
		"!": "Antonyms",
		"+": "Derivationally related forms",
		"&": "Similar to",
		"<": "Participle of verb",
		"\\": "Pertainyms",
		"=": "Attributes",
		"^": "Also see",
		";c": "Domain of synset - TOPIC",
		";r": "Domain of synset - REGION",
		";u": "Domain of synset - USAGE",
	}

	r = {
		"!": "Antonyms",
		"\\": "Derived from adjective",
		"+": "Derivationally related forms",
		";c": "Domain of synset - TOPIC",
		";r": "Domain of synset - REGION",
		";u": "Domain of synset - USAGE",
		"^": "Also see",
	}


class Pointer:
	def __init__(self, symbol: str, offset: str, pos: str, source_target: str) -> None:
		self.symbol = symbol
		self.offset = int(offset)
		self.pos = pos
		self.source_target = source_target
		self.source = int(source_target[:2], 16)
		self.target = int(source_target[2:], 16)

	def __repr__(self) -> str:
		return (
			f"Pointer({self.symbol!r}, {self.offset!r}, "
			f"{self.pos!r}, {self.source_target!r})"
		)


class WordNet:
	article_template = "<h1>%s</h1><span>%s</span>"
	synSetTypes = {
		"n": "n.",
		"v": "v.",
		"a": "adj.",
		"s": "adj. satellite",
		"r": "adv.",
	}

	file2pos = {
		"data.adj": ["a", "s"],
		"data.adv": ["r"],
		"data.noun": ["n"],
		"data.verb": ["v"],
	}

	def __init__(self, wordnetdir: str) -> None:
		self.wordnetdir = wordnetdir
		self.collector: "dict[str, list[str]]" = defaultdict(list)

	@staticmethod
	def iterlines(dict_dir: str) -> Iterator[str]:
		for name in os.listdir(dict_dir):
			if not name.startswith("data."):
				continue
			with open(os.path.join(dict_dir, name), encoding="utf-8") as f:
				for line in f:
					if not line.startswith("  "):
						yield line

	# PLR0912 Too many branches (16 > 12)
	def prepare(self) -> None:  # noqa: PLR0912
		synSetTypes = self.synSetTypes
		file2pos = self.file2pos

		dict_dir = self.wordnetdir

		files: dict[str, io.TextIOWrapper] = {}
		for name in os.listdir(dict_dir):
			if name.startswith("data.") and name in file2pos:
				f = open(os.path.join(dict_dir, name), encoding="utf-8")  # noqa: SIM115
				for key in file2pos[name]:
					files[key] = f

		def a(word: str) -> str:
			return f'<a href="{word}">{word}</a>'

		for index, line in enumerate(self.iterlines(dict_dir)):
			if index % 100 == 0 and index > 0:
				sys.stdout.write(".")
				sys.stdout.flush()
			if index % 5000 == 0 and index > 0:
				sys.stdout.write("\n")
				sys.stdout.flush()
			if not line or not line.strip():
				continue
			synset = SynSet(line)
			gloss_with_examples, _ = quotedTextPattern.subn(
				lambda x: f'<cite class="ex">{x.group(1)}</cite>',
				synset.gloss,
			)
			gloss_with_examples, _ = refPattern.subn(
				lambda x: a(x.group(1)),
				gloss_with_examples,
			)

			words = synset.words
			for index2, word in enumerate(words):
				# TODO: move this block to a func
				synonyms = ", ".join(a(w) for w in words if w != word)
				synonyms_str = (
					f'<br/><small class="co">Synonyms:</small> {synonyms}'
					if synonyms
					else ""
				)
				pointers = defaultdict(list)
				for pointer in synset.pointers:
					if (
						pointer.source
						and pointer.target
						and pointer.source - 1 != index2
					):
						continue
					symbol = pointer.symbol
					if symbol and symbol[:1] in {";", "-"}:
						continue
					try:
						symbol_desc = getattr(PointerSymbols, synset.ss_type)[symbol]
					except KeyError:
						log.warning(
							f"unknown pointer symbol {symbol} for {synset.ss_type} ",
						)
						symbol_desc = symbol

					data_file = files[pointer.pos]
					data_file.seek(pointer.offset)
					referenced_synset = SynSet(data_file.readline())
					if pointer.source == pointer.target == 0:
						pointers[symbol_desc] = [
							w for w in referenced_synset.words if w not in words
						]
					else:
						referenced_word = referenced_synset.words[pointer.target - 1]
						if referenced_word not in pointers[symbol_desc]:
							pointers[symbol_desc].append(referenced_word)

				pointers_str = "".join(
					[
						f'<br/><small class="co">{symbol_desc}:</small> '
						+ ", ".join(a(w) for w in referenced_words)
						for symbol_desc, referenced_words in pointers.items()
						if referenced_words
					],
				)
				self.collector[word].append(
					f'<i class="pos grammar">{synSetTypes[synset.ss_type]}</i>'
					f" {gloss_with_examples}{synonyms_str}{pointers_str}",
				)
		sys.stdout.write("\n")
		sys.stdout.flush()

	def process(self) -> "Iterator[tuple[str, str]]":
		article_template = self.article_template

		for title in self.collector:
			article_pieces = self.collector[title]
			article_pieces_count = len(article_pieces)
			text = None
			if article_pieces_count > 1:
				ol = ["<ol>"] + [f"<li>{ap}</li>" for ap in article_pieces] + ["</ol>"]
				text = article_template % (title, "".join(ol))
			elif article_pieces_count == 1:
				text = article_template % (title, article_pieces[0])

			if text:
				yield title, text


class Reader:
	def __init__(self, glos: "GlossaryType") -> None:
		self._glos = glos
		self._filename = ""
		self._wordCount = 0
		self.wordnet: "WordNet | None" = None

	def __len__(self) -> int:
		return self._wordCount

	def open(self, filename: str) -> None:
		self.wordnet = WordNet(filename)
		log.info("Running wordnet.prepare()")
		self.wordnet.prepare()

		# TODO: metadata

	def close(self) -> None:
		self.wordnet = None

	def __iter__(self) -> "Iterator[EntryType]":
		if self.wordnet is None:
			raise ValueError("self.wordnet is None")
		glos = self._glos
		for word, defi in self.wordnet.process():
			yield glos.newEntry(word, defi)



# -*- coding: utf-8 -*-
# glossary_utils.py
#
# Copyright © 2008-2022 Saeed Rasooli <saeed.gnu@gmail.com> (ilius)
# This file is part of PyGlossary project, https://github.com/ilius/pyglossary
#
# This program is a free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program. Or on Debian systems, from /usr/share/common-licenses/GPL
# If not, see <http://www.gnu.org/licenses/gpl.txt>.

import logging
from os.path import (
	splitext,
)

from .compression import (
	stdCompressions,
)

__all__ = ["splitFilenameExt"]

log = logging.getLogger("pyglossary")

MAX_EXT_LEN = 4  # FIXME


class Error(Exception):
	pass


class ReadError(Error):
	pass


class WriteError(Error):
	pass


def splitFilenameExt(
	filename: str = "",
) -> "tuple[str, str, str, str]":
	"""Return (filenameNoExt, filename, ext, compression)."""
	compression = ""
	filenameNoExt, ext = splitext(filename)
	ext = ext.lower()

	if not ext and len(filenameNoExt) <= MAX_EXT_LEN:
		filenameNoExt, ext = "", filenameNoExt

	if not ext:
		return filename, filename, "", ""

	if ext[1:] in {*stdCompressions, "zip", "dz"}:
		compression = ext[1:]
		filename = filenameNoExt
		filenameNoExt, ext = splitext(filename)
		ext = ext.lower()

	return filenameNoExt, filename, ext, compression



import sys
from itertools import (
	repeat,
	takewhile,
)

__all__ = ["fileCountLines"]


def fileCountLines(filename: str, newline: bytes = b"\n") -> int:
	with open(filename, "rb") as _file:
		bufgen = takewhile(
			lambda x: x,  # predicate
			(_file.read(1024 * 1024) for _ in repeat(None)),  # iterable
		)
		return sum(buf.count(newline) for buf in bufgen if buf)


if __name__ == "__main__":
	for filename in sys.argv[1:]:
		print(fileCountLines(filename), filename)  # noqa: T201



# -*- coding: utf-8 -*-
# mypy: ignore-errors
#
# Copyright © 2016-2017 Saeed Rasooli <saeed.gnu@gmail.com> (ilius)
#
# This program is a free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3, or (at your option)
# any later version.
#
# You can get a copy of GNU General Public License along this program
# But you can always get it from http://www.gnu.org/licenses/gpl.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.


from . import gdk, gtk
from .utils import imageFromFile

__all__ = ["ResizeButton"]


class ResizeButton(gtk.EventBox):
	def __init__(self, win, edge=gdk.WindowEdge.SOUTH_EAST) -> None:
		gtk.EventBox.__init__(self)
		self.win = win
		self.edge = edge
		###
		self.image = imageFromFile("resize.png")
		self.add(self.image)
		self.connect("button-press-event", self.buttonPress)

	def buttonPress(self, _obj, gevent):
		self.win.begin_resize_drag(
			self.edge,
			gevent.button,
			int(gevent.x_root),
			int(gevent.y_root),
			gevent.time,
		)



# -*- coding: utf-8 -*-

import os
import re
import shutil
from os.path import isfile, splitext
from typing import TYPE_CHECKING

if TYPE_CHECKING:
	from collections.abc import Generator, Iterator

	from pyglossary import slob

from pyglossary.core import cacheDir, log, pip
from pyglossary.glossary_types import EntryType, GlossaryType
from pyglossary.option import (
	BoolOption,
	FileSizeOption,
	IntOption,
	Option,
	StrOption,
)

__all__ = [
	"Reader",
	"Writer",
	"description",
	"enable",
	"extensionCreate",
	"extensions",
	"format",
	"kind",
	"lname",
	"optionsProp",
	"singleFile",
	"website",
	"wiki",
]

enable = True
lname = "aard2_slob"
format = "Aard2Slob"
description = "Aard 2 (.slob)"
extensions = (".slob",)
extensionCreate = ".slob"
singleFile = True
kind = "binary"
wiki = "https://github.com/itkach/slob/wiki"
website = (
	"http://aarddict.org/",
	"aarddict.org",
)
optionsProp: "dict[str, Option]" = {
	"compression": StrOption(
		values=["", "bz2", "zlib", "lzma2"],
		comment="Compression Algorithm",
	),
	"content_type": StrOption(
		customValue=True,
		values=[
			"text/plain; charset=utf-8",
			"text/html; charset=utf-8",
		],
		comment="Content Type",
	),
	# "encoding": EncodingOption(),
	"file_size_approx": FileSizeOption(
		comment="split up by given approximate file size\nexamples: 100m, 1g",
	),
	"file_size_approx_check_num_entries": IntOption(
		comment="for file_size_approx, check every `[?]` entries",
	),
	"separate_alternates": BoolOption(
		comment="add alternate headwords as separate entries to slob",
	),
	"word_title": BoolOption(
		comment="add headwords title to beginning of definition",
	),
	"version_info": BoolOption(
		comment="add version info tags to slob file",
	),
	"audio_goldendict": BoolOption(
		comment="Convert audio links for GoldenDict (desktop)",
	),
}

extraDocs = [
	(
		"PyICU",
		"See [doc/pyicu.md](./doc/pyicu.md) file for more detailed"
		" instructions on how to install PyICU.",
	),
]

t_created_at = "created.at"
t_label = "label"
t_created_by = "created.by"
t_copyright = "copyright"
t_license_name = "license.name"
t_license_url = "license.url"
t_uri = "uri"
t_edition = "edition"

supported_tags = {
	t_label,
	t_created_at,
	t_created_by,
	t_copyright,
	t_uri,
	t_edition,
}


class Reader:
	depends = {
		"icu": "PyICU",  # >=1.5
	}

	def __init__(self, glos: "GlossaryType") -> None:
		self._glos = glos
		self._clear()
		self._re_bword = re.compile(
			"(<a href=[^<>]+?>)",
			re.IGNORECASE,
		)

	def close(self) -> None:
		if self._slobObj is not None:
			self._slobObj.close()
		self._clear()

	def _clear(self) -> None:
		self._filename = ""
		self._slobObj: "slob.Slob | None" = None

	# TODO: PLR0912 Too many branches (13 > 12)
	def open(self, filename: str) -> None:  # noqa: PLR0912
		try:
			import icu  # type: ignore # noqa: F401
		except ModuleNotFoundError as e:
			e.msg += f", run `{pip} install PyICU` to install"
			raise e
		from pyglossary import slob

		self._filename = filename
		self._slobObj = slob.open(filename)
		tags = dict(self._slobObj.tags.items())

		if t_label in tags:
			self._glos.setInfo("name", tags[t_label])

		if t_created_at in tags:
			self._glos.setInfo("creationTime", tags[t_created_at])

		if t_created_by in tags:
			self._glos.setInfo("author", tags[t_created_by])

		copyrightLines = []
		for key in (t_copyright, t_license_name, t_license_url):
			try:
				value = tags.pop(key)
			except KeyError:
				continue
			copyrightLines.append(value)
		if copyrightLines:
			self._glos.setInfo("copyright", "\n".join(copyrightLines))

		if t_uri in tags:
			self._glos.setInfo("website", tags[t_uri])

		if t_edition in tags:
			self._glos.setInfo("edition", tags[t_edition])

		for key, value in tags.items():
			if key in supported_tags:
				continue
			self._glos.setInfo(f"slob.{key}", value)

	def __len__(self) -> int:
		if self._slobObj is None:
			log.error("called len() on a reader which is not open")
			return 0
		return len(self._slobObj)

	@staticmethod
	def _href_sub(m: "re.Match") -> str:
		st = m.group(0)
		if "//" in st:
			return st
		return st.replace('href="', 'href="bword://').replace(
			"href='",
			"href='bword://",
		)

	def __iter__(self) -> "Iterator[EntryType | None]":
		from pyglossary.slob import MIME_HTML, MIME_TEXT

		if self._slobObj is None:
			raise RuntimeError("iterating over a reader while it's not open")

		slobObj = self._slobObj
		blobSet = set()

		# slob library gives duplicate blobs when iterating over slobObj
		# even keeping the last id is not enough, since duplicate blobs
		# are not all consecutive. so we have to keep a set of blob IDs

		for blob in slobObj:
			_id = blob.identity
			if _id in blobSet:
				yield None  # update progressbar
				continue
			blobSet.add(_id)

			# blob.key is str, blob.content is bytes
			word = blob.key

			ctype = blob.content_type.split(";")[0]
			if ctype not in {MIME_HTML, MIME_TEXT}:
				log.debug(f"unknown {blob.content_type=} in {word=}")
				word = word.removeprefix("~/")
				yield self._glos.newDataEntry(word, blob.content)
				continue
			defiFormat = ""
			if ctype == MIME_HTML:
				defiFormat = "h"
			elif ctype == MIME_TEXT:
				defiFormat = "m"

			defi = blob.content.decode("utf-8")
			defi = self._re_bword.sub(self._href_sub, defi)
			yield self._glos.newEntry(word, defi, defiFormat=defiFormat)


class Writer:
	depends = {
		"icu": "PyICU",
	}

	_compression: str = "zlib"
	_content_type: str = ""
	_file_size_approx: int = 0
	_file_size_approx_check_num_entries = 100
	_separate_alternates: bool = False
	_word_title: bool = False
	_version_info: bool = False

	_audio_goldendict: bool = False

	resourceMimeTypes = {
		"png": "image/png",
		"jpeg": "image/jpeg",
		"jpg": "image/jpeg",
		"gif": "image/gif",
		"svg": "image/svg+xml",
		"webp": "image/webp",
		"tiff": "image/tiff",
		"tif": "image/tiff",
		"bmp": "image/bmp",
		"css": "text/css",
		"js": "application/javascript",
		"json": "application/json",
		"woff": "application/font-woff",
		"woff2": "application/font-woff2",
		"ttf": "application/x-font-ttf",
		"otf": "application/x-font-opentype",
		"mp3": "audio/mpeg",
		"ogg": "audio/ogg",
		"spx": "audio/x-speex",
		"wav": "audio/wav",
		"ini": "text/plain",
		# "application/octet-stream+xapian",
		"eot": "application/vnd.ms-fontobject",
		"pdf": "application/pdf",
		"mp4": "video/mp4",
	}

	def __init__(self, glos: GlossaryType) -> None:
		self._glos = glos
		self._filename = ""
		self._resPrefix = ""
		self._slobWriter: "slob.Writer | None" = None

	@staticmethod
	def _slobObserver(
		event: "slob.WriterEvent",  # noqa: F401, F821
	) -> None:
		log.debug(f"slob: {event.name}{': ' + event.data if event.data else ''}")

	def _open(self, filename: str, namePostfix: str) -> "slob.Writer":
		from pyglossary import slob

		if isfile(filename):
			shutil.move(filename, f"{filename}.bak")
			log.warning(f"renamed existing {filename!r} to {filename + '.bak'!r}")
		self._slobWriter = slobWriter = slob.Writer(
			filename,
			observer=self._slobObserver,
			workdir=cacheDir,
			compression=self._compression,
			version_info=self._version_info,
		)
		slobWriter.tag("label", self._glos.getInfo("name") + namePostfix)
		return slobWriter

	def open(self, filename: str) -> None:
		try:
			import icu  # noqa: F401
		except ModuleNotFoundError as e:
			e.msg += f", run `{pip} install PyICU` to install"
			raise e
		if isfile(filename):
			raise OSError(f"File '{filename}' already exists")
		namePostfix = ""
		if self._file_size_approx > 0:
			namePostfix = " (part 1)"
		self._open(filename, namePostfix)
		self._filename = filename

	def finish(self) -> None:
		from time import perf_counter

		self._filename = ""
		if self._slobWriter is None:
			return
		log.info("Finalizing slob file...")
		t0 = perf_counter()
		self._slobWriter.finalize()
		log.info(f"Finalizing slob file took {perf_counter() - t0:.1f} seconds")
		self._slobWriter = None

	def addDataEntry(self, entry: "EntryType") -> None:
		slobWriter = self._slobWriter
		if slobWriter is None:
			raise ValueError("slobWriter is None")
		rel_path = entry.s_word
		_, ext = splitext(rel_path)
		ext = ext.lstrip(os.path.extsep).lower()
		content_type = self.resourceMimeTypes.get(ext)
		if not content_type:
			log.error(f"Aard2 slob: unknown content type for {rel_path!r}")
			return
		content = entry.data
		key = self._resPrefix + rel_path
		try:
			key.encode(slobWriter.encoding)
		except UnicodeEncodeError:
			log.error(f"Failed to add, broken unicode in key: {key!a}")
			return
		slobWriter.add(content, key, content_type=content_type)

	def addEntry(self, entry: "EntryType") -> None:
		words = entry.l_word
		b_defi = entry.defi.encode("utf-8")
		_ctype = self._content_type
		writer = self._slobWriter
		if writer is None:
			raise ValueError("slobWriter is None")

		entry.detectDefiFormat()
		defiFormat = entry.defiFormat

		if self._word_title and defiFormat in {"h", "m"}:
			if defiFormat == "m":
				defiFormat = "h"
			title = self._glos.wordTitleStr(
				words[0],
			)
			b_defi = title.encode("utf-8") + b_defi

		if defiFormat == "h":
			b_defi = b_defi.replace(b'"bword://', b'"')
			b_defi = b_defi.replace(b"'bword://", b"'")

			if not self._audio_goldendict:
				b_defi = b_defi.replace(
					b"""href="sound://""",
					b'''onclick="new Audio(this.href).play(); return false;" href="''',
				)
				b_defi = b_defi.replace(
					b"""href='sound://""",
					b"""onclick="new Audio(this.href).play(); return false;" href='""",
				)
				b_defi = b_defi.replace(b"""<img src="/""", b'''<img src="''')
				b_defi = b_defi.replace(b"""<img src='""", b"""<img src='""")
				b_defi = b_defi.replace(b"""<img src="file:///""", b'''<img src="''')
				b_defi = b_defi.replace(b"""<img src='file:///""", b"""<img src='""")

		if not _ctype:
			if defiFormat == "h":
				_ctype = "text/html; charset=utf-8"
			elif defiFormat == "m":
				_ctype = "text/plain; charset=utf-8"
			else:
				_ctype = "text/plain; charset=utf-8"

		if not self._separate_alternates:
			writer.add(
				b_defi,
				*tuple(words),
				content_type=_ctype,
			)
			return

		headword, *alts = words
		writer.add(
			b_defi,
			headword,
			content_type=_ctype,
		)
		for alt in alts:
			writer.add(
				b_defi,
				f"{alt}, {headword}",
				content_type=_ctype,
			)

	def write(self) -> "Generator[None, EntryType, None]":
		slobWriter = self._slobWriter
		if slobWriter is None:
			raise ValueError("slobWriter is None")
		file_size_approx = int(self._file_size_approx * 0.95)
		entryCount = 0
		sumBlobSize = 0
		fileIndex = 0
		filenameNoExt, _ = splitext(self._filename)
		while True:
			entry = yield
			if entry is None:
				break

			if entry.isData():
				self.addDataEntry(entry)
			else:
				self.addEntry(entry)

			if file_size_approx <= 0:
				continue

			# handle file_size_approx
			check_every = self._file_size_approx_check_num_entries
			entryCount += 1
			if entryCount % check_every == 0:
				sumBlobSize = slobWriter.size_data()
				if sumBlobSize >= file_size_approx:
					slobWriter.finalize()
					fileIndex += 1
					slobWriter = self._open(
						f"{filenameNoExt}.{fileIndex}.slob",
						f" (part {fileIndex + 1})",
					)
					sumBlobSize = 0
					entryCount = 0



# -*- coding: utf-8 -*-
#
# Copyright © 2016-2019 Saeed Rasooli <saeed.gnu@gmail.com> (ilius)
# Copyright © 2016 ivan tkachenko me@ratijas.tk
# Copyright © 2012-2015 Xiaoqiang Wang <xiaoqiangwang AT gmail DOT com>
#
# This program is a free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3 of the License.
#
# You can get a copy of GNU General Public License along this program
# But you can always get it from http://www.gnu.org/licenses/gpl.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.

# FIXME:
# MDX-specific parts should be isolated and moved to MDX Reader
# and parts that are specific to one glossary
# (like Oxford_Advanced_English-Chinese_Dictionary_9th_Edition.mdx)
# should be moved to separate modules (like content processors) and enabled
# per-glossary (by title or something else)

import logging
import re
from typing import Any
from xml.sax.saxutils import quoteattr, unescape

from pyglossary.text_utils import toStr

__all__ = ["prepare_content"]

log = logging.getLogger("pyglossary")


re_brhr = re.compile("<(BR|HR)>", re.IGNORECASE)
re_nonprintable = re.compile("[\x00-\x07\x0e-\x1f]")
re_img = re.compile("<IMG (.*?)>", re.IGNORECASE)

re_div_margin_em = re.compile(r'<div style="margin-left:(\d)em">')
sub_div_margin_em = r'<div class="m\1">'

re_div_margin_em_ex = re.compile(
	r'<div class="ex" style="margin-left:(\d)em;color:steelblue">',
)
sub_div_margin_em_ex = r'<div class="m\1 ex">'

re_href = re.compile(r"""href=(["'])(.*?)\1""")

re_margin = re.compile(r"margin-left:(\d)em")


def prepare_content(
	title: "str | None",
	body: str,
	BeautifulSoup: "Any",
) -> str:
	# heavily integrated with output of dsl reader plugin!
	# and with xdxf also.
	""":param title: str | None"""
	# class="sec" => d:priority="2"
	# style="color:steelblue" => class="ex"
	# class="p" style="color:green" => class="p"
	# style="color:green" => class="c"
	# style="margin-left:{}em" => class="m{}"
	# <s> => <del>

	# xhtml is strict
	if BeautifulSoup:
		content = prepare_content_with_soup(title, body, BeautifulSoup)
	else:
		content = prepare_content_without_soup(title, body)

	content = content.replace("&nbsp;", "&#160;")
	content = re_nonprintable.sub("", content)
	return content  # noqa: RET504


def prepare_content_without_soup(
	title: "str | None",
	body: str,
) -> str:
	# somewhat analogue to what BeautifulSoup suppose to do
	body = re_div_margin_em.sub(sub_div_margin_em, body)
	body = re_div_margin_em_ex.sub(sub_div_margin_em_ex, body)
	body = re_href.sub(href_sub, body)

	body = (
		body.replace(
			'<i style="color:green">',
			'<i class="c">',
		)
		.replace(
			'<i class="p" style="color:green">',
			'<i class="p">',
		)
		.replace(
			'<span class="ex" style="color:steelblue">',
			'<span class="ex">',
		)
		.replace(
			'<span class="sec ex" style="color:steelblue">',
			'<span class="sec ex">',
		)
		.replace("<u>", '<span class="u">')
		.replace("</u>", "</span>")
		.replace("<s>", "<del>")
		.replace("</s>", "</del>")
	)

	# nice header to display
	content = f"<h1>{title}</h1>{body}" if title else body
	content = re_brhr.sub(r"<\g<1> />", content)
	content = re_img.sub(r"<img \g<1>/>", content)
	return content  # noqa: RET504


def _prepare_href(tag) -> None:
	href = tag["href"]
	href = cleanup_link_target(href)

	if href.startswith("sound:"):
		fix_sound_link(href, tag)

	elif href.startswith(("phonetics", "help:phonetics")):
		# for oxford9
		log.debug(f"phonetics: {tag=}")
		if tag.audio and "name" in tag.audio.attrs:
			tag["onmousedown"] = "this.lastChild.play(); return false;"
			src_name = tag.audio["name"].replace("#", "_")
			tag.audio["src"] = f"{src_name}.mp3"

	elif not link_is_url(href):
		tag["href"] = f"x-dictionary:d:{href}"


def _prepare_onclick(soup) -> None:
	for thumb in soup.find_all("div", "pic_thumb"):
		thumb["onclick"] = (
			'this.setAttribute("style", "display:none"); '
			'this.nextElementSibling.setAttribute("style", "display:block")'
		)

	for pic in soup.find_all("div", "big_pic"):
		pic["onclick"] = (
			'this.setAttribute("style", "display:none"), '
			'this.previousElementSibling.setAttribute("style", "display:block")'
		)

	# to unfold(expand) and fold(collapse) blocks
	for pos in soup.find_all("pos", onclick="toggle_infl(this)"):
		# TODO: simplify this!
		pos["onclick"] = (
			r"var e = this.parentElement.parentElement.parentElement"
			r'.querySelector("res-g vp-gs"); style = window.'
			r"getComputedStyle(e), display = style.getPropertyValue"
			r'("display"), "none" === e.style.display || "none" === display'
			r' ? e.style.display = "block" : e.style.display = "none", '
			r"this.className.match(/(?:^|\s)Clicked(?!\S)/) ? this."
			r"className = this.className.replace("
			r'/(?:^|\s)Clicked(?!\S)/g, "") : this.setAttribute('
			r'"class", "Clicked")'
		)


# TODO: PLR0912 Too many branches (18 > 12)
def prepare_content_with_soup(  # noqa: PLR0912
	title: "str | None",
	body: str,
	BeautifulSoup: "Any",
) -> str:
	soup = BeautifulSoup.BeautifulSoup(body, features="lxml")
	# difference between "lxml" and "html.parser"
	if soup.body:
		soup = soup.body

	for tag in soup(class_="sec"):
		tag["class"].remove("sec")
		if not tag["class"]:
			del tag["class"]
		tag["d:priority"] = "2"

	for tag in soup(lambda x: "color:steelblue" in x.get("style", "")):
		remove_style(tag, "color:steelblue")
		if "ex" not in tag.get("class", []):
			tag["class"] = tag.get("class", []) + ["ex"]

	for tag in soup(is_green):
		remove_style(tag, "color:green")
		if "p" not in tag.get("class", ""):
			tag["class"] = tag.get("class", []) + ["c"]

	for tag in soup(True):
		if "style" in tag.attrs:
			m = re_margin.search(tag["style"])
			if m:
				remove_style(tag, m.group(0))
				tag["class"] = tag.get("class", []) + ["m" + m.group(1)]

	for tag in soup(lambda x: "xhtml:" in x.name):
		old_tag_name = tag.name
		tag.name = old_tag_name[len("xhtml:") :]
		if tag.string:
			tag.string = f"{tag.string} "

	for tag in soup.select("[href]"):
		_prepare_href(tag)

	_prepare_onclick(soup)

	for tag in soup.select("[src]"):
		src = tag["src"]
		if src.startswith("/"):
			tag["src"] = src[1:]
	for tag in soup("u"):
		tag.name = "span"
		tag["class"] = tag.get("class", []) + ["u"]
	for tag in soup("s"):
		tag.name = "del"

	if title and "<h" not in body:
		h1 = BeautifulSoup.Tag(name="h1")
		h1.string = title
		soup.insert(0, h1)

	# hence the name BeautifulSoup
	# soup.insert(0, head)
	return toStr(soup.encode_contents())


def cleanup_link_target(href: str) -> str:
	return href.removeprefix("bword://")


def href_sub(x: "re.Match") -> str:
	href = x.groups()[1]
	if href.startswith("http"):
		return x.group()

	href = cleanup_link_target(href)

	return "href=" + quoteattr(
		"x-dictionary:d:"
		+ unescape(
			href,
			{"&quot;": '"'},
		),
	)


def is_green(x: dict) -> bool:
	return "color:green" in x.get("style", "")


def remove_style(tag: dict, line: str) -> None:
	s = "".join(tag["style"].replace(line, "").split(";"))
	if s:
		tag["style"] = s
	else:
		del tag["style"]


def fix_sound_link(href: str, tag: "dict[str, Any]") -> None:
	tag["href"] = f'javascript:new Audio("{href[len("sound://"):]}").play();'


def link_is_url(href: str) -> bool:
	for prefix in (
		"http:",
		"https:",
		"addexample:",
		"addid:",
		"addpv:",
		"help:",
		"helpg:",
		"helpp:",
		"helpr:",
		"helpxr:",
		"xi:",
		"xid:",
		"xp:",
		"sd:",
		"#",
	):
		if href.startswith(prefix):
			return True
	return False



import argparse


class StoreConstAction(argparse.Action):
	def __init__(
		self,
		option_strings: "list[str]",
		same_dest: str = "",
		const_value: "bool | None" = None,
		nargs: int = 0,
		**kwargs,
	) -> None:
		if isinstance(option_strings, str):
			option_strings = [option_strings]
		argparse.Action.__init__(
			self,
			option_strings=option_strings,
			nargs=nargs,
			**kwargs,
		)
		self.same_dest = same_dest
		self.const_value = const_value

	def __call__(  # noqa: PLR0913
		self,
		parser: "argparse.ArgumentParser | None" = None,
		namespace: "argparse.Namespace | None" = None,
		values: list | None = None,  # noqa: ARG002
		option_strings: list[str] | None = None,  # noqa: ARG002
		required: bool = False,  # noqa: ARG002
		dest: str | None = None,
	) -> "StoreConstAction":
		if not parser:
			return self
		dest = self.dest
		if getattr(namespace, dest) is not None:
			flag = self.option_strings[0]
			if getattr(namespace, dest) == self.const_value:
				parser.error(f"multiple {flag} options")
			else:
				parser.error(f"conflicting options: {self.same_dest} and {flag}")
		setattr(namespace, dest, self.const_value)
		return self



from typing import TYPE_CHECKING

from .core import log

if TYPE_CHECKING:
	from .ui_type import UIType

__all__ = ["GlossaryProgress"]


class GlossaryProgress:
	def __init__(
		self,
		ui: "UIType | None" = None,  # noqa: F821
	) -> None:
		self._ui = ui
		self._progressbar = True

	def clear(self) -> None:
		self._progressbar = True

	@property
	def progressbar(self) -> bool:
		return self._ui is not None and self._progressbar

	@progressbar.setter
	def progressbar(self, enabled: bool) -> None:
		self._progressbar = enabled

	def progressInit(
		self,
		*args,  # noqa: ANN003
	) -> None:
		if self._ui and self._progressbar:
			self._ui.progressInit(*args)

	def progress(self, pos: int, total: int, unit: str = "entries") -> None:
		if total == 0:
			log.warning(f"{pos=}, {total=}")
			return
		if self._ui is None:
			return
		self._ui.progress(
			min(pos + 1, total) / total,
			f"{pos:,} / {total:,} {unit}",
		)

	def progressEnd(self) -> None:
		if self._ui and self._progressbar:
			self._ui.progressEnd()



import inspect
import logging
import os
import platform
import sys
import traceback
from os.path import (
	abspath,
	dirname,
	exists,
	isdir,
	isfile,
	join,
)
from typing import TYPE_CHECKING, cast

if TYPE_CHECKING:
	from collections.abc import Callable
	from types import TracebackType
	from typing import (
		Any,
		TypeAlias,
	)

	ExcInfoType: TypeAlias = (
		"tuple[type[BaseException], BaseException, TracebackType]"
		" | tuple[None, None, None]"
	)


__all__ = [
	"TRACE",
	"VERSION",
	"StdLogHandler",
	"appResDir",
	"cacheDir",
	"checkCreateConfDir",
	"confDir",
	"confJsonFile",
	"dataDir",
	"format_exception",
	"getDataDir",
	"homeDir",
	"homePage",
	"isDebug",
	"log",
	"noColor",
	"pip",
	"pluginsDir",
	"rootConfJsonFile",
	"rootDir",
	"sysName",
	"tmpDir",
	"trace",
	"uiDir",
	"userPluginsDir",
]


VERSION = "4.7.1"

homePage = "https://github.com/ilius/pyglossary"


TRACE = 5
logging.addLevelName(TRACE, "TRACE")
noColor = False


def trace(log: logging.Logger, msg: str) -> None:
	func = getattr(log, "trace", None)
	if func is None:
		log.error(f"Logger {log} has no 'trace' method")
		return
	func(msg)


class _Formatter(logging.Formatter):
	def __init__(self, *args, **kwargs) -> None:  # noqa: ANN101
		logging.Formatter.__init__(self, *args, **kwargs)
		self.fill: "Callable[[str], str] | None" = None

	def formatMessage(
		self,
		record: "logging.LogRecord",
	) -> str:
		msg = logging.Formatter.formatMessage(self, record)
		if self.fill is not None:
			msg = self.fill(msg)
		return msg  # noqa: RET504


class _MyLogger(logging.Logger):
	levelsByVerbosity = (
		logging.CRITICAL,
		logging.ERROR,
		logging.WARNING,
		logging.INFO,
		logging.DEBUG,
		TRACE,
		logging.NOTSET,
	)
	levelNamesCap = (
		"Critical",
		"Error",
		"Warning",
		"Info",
		"Debug",
		"Trace",
		"All",  # "Not-Set",
	)

	def __init__(self, *args) -> None:  # noqa: ANN101
		logging.Logger.__init__(self, *args)
		self._verbosity = 3
		self._timeEnable = False

	def setVerbosity(self, verbosity: int) -> None:
		self.setLevel(self.levelsByVerbosity[verbosity])
		self._verbosity = verbosity

	def getVerbosity(self) -> int:
		return self._verbosity

	def trace(self, msg: str) -> None:
		self.log(TRACE, msg)

	def pretty(self, data: "Any", header: str = "") -> None:
		from pprint import pformat

		self.debug(header + pformat(data))

	def newFormatter(self) -> _Formatter:
		timeEnable = self._timeEnable
		if timeEnable:
			fmt = "%(asctime)s [%(levelname)s] %(message)s"
		else:
			fmt = "[%(levelname)s] %(message)s"
		return _Formatter(fmt)

	def setTimeEnable(self, timeEnable: bool) -> None:
		self._timeEnable = timeEnable
		formatter = self.newFormatter()
		for handler in self.handlers:
			handler.setFormatter(formatter)

	def addHandler(self, hdlr: "logging.Handler") -> None:
		# if want to add separate format (new config keys and flags) for ui_gtk
		# and ui_tk, you need to remove this function and run handler.setFormatter
		# in ui_gtk and ui_tk
		logging.Logger.addHandler(self, hdlr)
		hdlr.setFormatter(self.newFormatter())


def _formatVarDict(
	dct: "dict[str, Any]",
	indent: int = 4,
	max_width: int = 80,
) -> str:
	lines = []
	pre = " " * indent
	for key, value in dct.items():
		line = pre + key + " = " + repr(value)
		if len(line) > max_width:
			line = line[: max_width - 3] + "..."
			try:
				value_len = len(value)
			except TypeError:
				pass
			else:
				line += f"\n{pre}len({key}) = {value_len}"
		lines.append(line)
	return "\n".join(lines)


def format_exception(
	exc_info: "ExcInfoType | None" = None,
	add_locals: bool = False,
	add_globals: bool = False,
) -> str:
	if exc_info is None:
		exc_info = sys.exc_info()
	_type, value, tback = exc_info
	text = "".join(traceback.format_exception(_type, value, tback))

	if tback is None:
		return text

	if add_locals or add_globals:
		try:
			frame = inspect.getinnerframes(tback, context=0)[-1][0]
		except IndexError:
			pass
		else:
			if add_locals:
				text += f"Traceback locals:\n{_formatVarDict(frame.f_locals)}\n"
			if add_globals:
				text += f"Traceback globals:\n{_formatVarDict(frame.f_globals)}\n"

	return text


class StdLogHandler(logging.Handler):
	colorsConfig = {
		"CRITICAL": ("color.cmd.critical", 196),
		"ERROR": ("color.cmd.error", 1),
		"WARNING": ("color.cmd.warning", 208),
	}
	# 1: dark red (like 31m), 196: real red, 9: light red
	# 15: white, 229: light yellow (#ffffaf), 226: real yellow (#ffff00)

	def __init__(self, noColor: bool = False) -> None:
		logging.Handler.__init__(self)
		self.set_name("std")
		self.noColor = noColor
		self.config: "dict[str, Any]" = {}

	@property
	def endFormat(self) -> str:
		if self.noColor:
			return ""
		return "\x1b[0;0;0m"

	def emit(self, record: logging.LogRecord) -> None:
		msg = ""
		if record.getMessage():
			msg = self.format(record)
		###
		if record.exc_info:
			_type, value, tback = record.exc_info
			if _type and tback and value:  # to fix mypy error
				tback_text = format_exception(
					exc_info=(_type, value, tback),
					add_locals=(log.level <= logging.DEBUG),
					add_globals=False,
				)
				if not msg:
					msg = "unhandled exception:"
				msg += "\n"
				msg += tback_text
		###
		levelname = record.levelname

		fp = sys.stderr if levelname in {"CRITICAL", "ERROR"} else sys.stdout

		if not self.noColor and levelname in self.colorsConfig:
			key, default = self.colorsConfig[levelname]
			colorCode = self.config.get(key, default)
			startColor = f"\x1b[38;5;{colorCode}m"
			msg = startColor + msg + self.endFormat

		###
		if fp is None:
			print(f"fp=None, levelname={record.levelname}")  # noqa: T201
			print(msg)  # noqa: T201
			return
		fp.write(msg + "\n")
		fp.flush()


def checkCreateConfDir() -> None:
	if not isdir(confDir):
		if exists(confDir):  # file, or anything other than directory
			os.rename(confDir, confDir + ".bak")  # we do not import old config
		os.mkdir(confDir)
	if not exists(userPluginsDir):
		try:
			os.mkdir(userPluginsDir)
		except Exception as e:
			log.warning(f"failed to create user plugins directory: {e}")
	if not isfile(confJsonFile):
		with (
			open(rootConfJsonFile, encoding="utf-8") as srcF,
			open(confJsonFile, "w", encoding="utf-8") as usrF,
		):
			usrF.write(srcF.read())


def _in_virtualenv() -> bool:
	if hasattr(sys, "real_prefix"):
		return True
	return hasattr(sys, "base_prefix") and sys.base_prefix != sys.prefix


def getDataDir() -> str:
	if _in_virtualenv():
		pass  # TODO
		# print(f"prefix={sys.prefix}, base_prefix={sys.base_prefix}")
		# return join(
		# 	dirname(dirname(dirname(rootDir))),
		# 	os.getenv("VIRTUAL_ENV"), "share", "pyglossary",
		# )

	if not rootDir.endswith(("dist-packages", "site-packages")):
		return rootDir

	parent3 = dirname(dirname(dirname(rootDir)))
	if os.sep == "/":
		return join(parent3, "share", "pyglossary")

	_dir = join(
		parent3,
		f"Python{sys.version_info.major}{sys.version_info.minor}",
		"share",
		"pyglossary",
	)
	if isdir(_dir):
		return _dir

	_dir = join(parent3, "Python3", "share", "pyglossary")
	if isdir(_dir):
		return _dir

	_dir = join(parent3, "Python", "share", "pyglossary")
	if isdir(_dir):
		return _dir

	_dir = join(sys.prefix, "share", "pyglossary")
	if isdir(_dir):
		return _dir

	if CONDA_PREFIX := os.getenv("CONDA_PREFIX"):
		_dir = join(CONDA_PREFIX, "share", "pyglossary")
		if isdir(_dir):
			return _dir

	raise OSError("failed to detect dataDir")


# __________________________________________________________________________ #


logging.setLoggerClass(_MyLogger)
log = cast(_MyLogger, logging.getLogger("pyglossary"))


def isDebug() -> bool:
	return log.getVerbosity() >= 4  # noqa: PLR2004


if os.sep == "\\":

	def _windows_show_exception(
		_type: "type[BaseException]",
		exc: "BaseException",
		tback: "TracebackType | None",
	) -> None:
		if not (_type and exc and tback):
			return
		import ctypes

		msg = format_exception(
			exc_info=(_type, exc, tback),
			add_locals=(log.level <= logging.DEBUG),
			add_globals=False,
		)
		log.critical(msg)
		ctypes.windll.user32.MessageBoxW(0, msg, "PyGlossary Error", 0)  # type: ignore

	sys.excepthook = _windows_show_exception

else:

	def _unix_show_exception(
		_type: "type[BaseException]",
		exc: "BaseException",
		tback: "TracebackType | None",
	) -> None:
		if not (_type and exc and tback):
			return
		log.critical(
			format_exception(
				exc_info=(_type, exc, tback),
				add_locals=(log.level <= logging.DEBUG),
				add_globals=False,
			),
		)

	sys.excepthook = _unix_show_exception

sysName = platform.system().lower()
# platform.system() is in	["Linux", "Windows", "Darwin", "FreeBSD"]
# sysName is in				["linux", "windows", "darwin', "freebsd"]


# can set env var WARNINGS to:
# "error", "ignore", "always", "default", "module", "once"
if WARNINGS := os.getenv("WARNINGS"):
	if WARNINGS in {"default", "error", "ignore", "always", "module", "once"}:
		import warnings

		warnings.filterwarnings(WARNINGS)  # type: ignore # noqa: PGH003
	else:
		log.error(f"invalid env var {WARNINGS = }")


if getattr(sys, "frozen", False):
	# PyInstaller frozen executable
	log.info(f"sys.frozen = {getattr(sys, 'frozen', False)}")
	rootDir = dirname(sys.executable)
	uiDir = join(rootDir, "pyglossary", "ui")
else:
	_srcDir = dirname(abspath(__file__))
	uiDir = join(_srcDir, "ui")
	rootDir = dirname(_srcDir)

dataDir = getDataDir()
appResDir = join(dataDir, "res")

if os.sep == "/":  # Operating system is Unix-Like
	homeDir = os.getenv("HOME", "/")
	tmpDir = os.getenv("TMPDIR", "/tmp")  # noqa: S108
	if sysName == "darwin":  # MacOS X
		_libDir = join(homeDir, "Library")
		confDir = join(_libDir, "Preferences", "PyGlossary")
		# or maybe: join(_libDir, "PyGlossary")
		# os.environ["OSTYPE"] == "darwin10.0"
		# os.environ["MACHTYPE"] == "x86_64-apple-darwin10.0"
		# platform.dist() == ("", "", "")
		# platform.release() == "10.3.0"
		cacheDir = join(_libDir, "Caches", "PyGlossary")
		pip = "pip3"
	else:  # GNU/Linux, Termux, FreeBSD, etc
		# should switch to "$XDG_CONFIG_HOME/pyglossary" in version 5.0.0
		# which generally means ~/.config/pyglossary
		confDir = join(homeDir, ".pyglossary")
		cacheDir = join(homeDir, ".cache", "pyglossary")
		pip = "pip3" if "/com.termux/" in homeDir else "sudo pip3"
elif os.sep == "\\":  # Operating system is Windows
	# FIXME: default values
	_HOMEDRIVE = os.getenv("HOMEDRIVE", "")
	_HOMEPATH = os.getenv("HOMEPATH", "")
	homeDir = join(_HOMEDRIVE, _HOMEPATH)
	tmpDir = os.getenv("TEMP", "")
	_appData = os.getenv("APPDATA", "")
	confDir = join(_appData, "PyGlossary")
	_localAppData = os.getenv("LOCALAPPDATA")
	if not _localAppData:
		# Windows Vista or older
		_localAppData = abspath(join(_appData, "..", "Local"))
	cacheDir = join(_localAppData, "PyGlossary", "Cache")
	pip = "pip3"
else:
	raise RuntimeError(
		f"Unknown path separator(os.sep=={os.sep!r}), unknown operating system!",
	)

pluginsDir = join(rootDir, "pyglossary", "plugins")
confJsonFile = join(confDir, "config.json")
rootConfJsonFile = join(dataDir, "config.json")
userPluginsDir = join(confDir, "plugins")



import typing
from collections.abc import (
	Callable,
	Iterator,  # noqa: TCH003
)

# -*- coding: utf-8 -*-
from typing import (
	TYPE_CHECKING,
	Any,
)

if TYPE_CHECKING:
	from collections import OrderedDict
	from typing import TypeAlias

	from .langs import Lang
	from .sort_keys import NamedSortKey

__all__ = [
	"Callable",
	"EntryListType",
	"EntryType",
	"GlossaryExtendedType",
	"GlossaryType",
	"RawEntryType",
]

MultiStr: "TypeAlias" = "str | list[str]"

# 3 different types in order:
# - compressed
# - uncompressed, without defiFormat
# - uncompressed, with defiFormat
RawEntryType: "TypeAlias" = (
	"bytes |tuple[list[str], bytes] |tuple[list[str], bytes, str]"
)


class EntryType(typing.Protocol):  # noqa: PLR0904
	def __init__(self) -> None: ...

	def isData(self) -> bool: ...

	def getFileName(self) -> str: ...

	@property
	def data(self) -> bytes: ...

	def size(self) -> int: ...

	def save(self, directory: str) -> str: ...

	@property
	def s_word(self) -> str: ...

	@property
	def l_word(self) -> "list[str]": ...

	@property
	def defi(self) -> str: ...

	@property
	def b_word(self) -> bytes: ...

	@property
	def b_defi(self) -> bytes: ...

	@property
	def defiFormat(self) -> str:
		# TODO: type: Literal["m", "h", "x", "b"]
		...

	@defiFormat.setter
	def defiFormat(self, defiFormat: str) -> None:
		# TODO: type: Literal["m", "h", "x", "b"]
		...

	def detectDefiFormat(self) -> None: ...

	def addAlt(self, alt: str) -> None: ...

	def editFuncWord(self, func: "Callable[[str], str]") -> None: ...

	def editFuncDefi(self, func: "Callable[[str], str]") -> None: ...

	def strip(self) -> None: ...

	def replaceInWord(self, source: str, target: str) -> None: ...

	def replaceInDefi(self, source: str, target: str) -> None: ...

	def replace(self, source: str, target: str) -> None: ...

	def byteProgress(self) -> "tuple[int, int] | None": ...

	def removeEmptyAndDuplicateAltWords(self) -> None: ...

	def stripFullHtml(self) -> "str | None": ...


class EntryListType(typing.Protocol):
	def __init__(
		self,
		entryToRaw: "Callable[[EntryType], RawEntryType]",
		entryFromRaw: "Callable[[RawEntryType], EntryType]",
	) -> None: ...

	@property
	def rawEntryCompress(self) -> bool: ...

	@rawEntryCompress.setter
	def rawEntryCompress(self, enable: bool) -> None: ...

	def append(self, entry: "EntryType") -> None: ...

	def clear(self) -> None: ...

	def __len__(self) -> int: ...

	def __iter__(self) -> "Iterator[EntryType]": ...

	def setSortKey(
		self,
		namedSortKey: "NamedSortKey",
		sortEncoding: "str | None",
		writeOptions: "dict[str, Any]",
	) -> None: ...

	def sort(self) -> None: ...

	def close(self) -> None: ...


class GlossaryType(typing.Protocol):  # noqa: PLR0904

	"""
	an abstract type class for Glossary class in plugins. it only
	contains methods and properties that might be used in plugins.
	"""

	def __iter__(self) -> "Iterator[EntryType]": ...

	def __len__(self) -> int: ...

	def setDefaultDefiFormat(self, defiFormat: str) -> None: ...

	def getDefaultDefiFormat(self) -> str: ...

	def collectDefiFormat(
		self,
		maxCount: int,
	) -> "dict[str, float] | None": ...

	def iterInfo(self) -> "Iterator[tuple[str, str]]": ...

	def getInfo(self, key: str) -> str: ...

	def setInfo(self, key: str, value: str) -> None: ...

	def getExtraInfos(self, excludeKeys: "list[str]") -> "OrderedDict": ...

	@property
	def author(self) -> str: ...

	@property
	def alts(self) -> bool: ...

	@property
	def filename(self) -> str: ...

	@property
	def tmpDataDir(self) -> str: ...

	@property
	def readOptions(self) -> dict | None: ...

	@property
	def sourceLang(self) -> "Lang | None": ...

	@property
	def targetLang(self) -> "Lang | None": ...

	@property
	def sourceLangName(self) -> str: ...

	@sourceLangName.setter
	def sourceLangName(self, langName: str) -> None: ...

	@property
	def targetLangName(self) -> str: ...

	@targetLangName.setter
	def targetLangName(self, langName: str) -> None: ...

	def titleTag(self, sample: str) -> str: ...

	def wordTitleStr(
		self,
		word: str,
		sample: str = "",
		_class: str = "",
	) -> str: ...

	def getConfig(self, name: str, default: "str | None") -> "str | None": ...

	def addEntry(self, entry: EntryType) -> None: ...

	def newEntry(
		self,
		word: "MultiStr",
		defi: str,
		defiFormat: str = "",
		byteProgress: "tuple[int, int] | None" = None,
	) -> EntryType: ...

	def newDataEntry(self, fname: str, data: bytes) -> EntryType: ...

	@property
	def rawEntryCompress(self) -> bool: ...

	def stripFullHtml(
		self,
		errorHandler: "Callable[[EntryType, str], None] | None" = None,
	) -> None: ...

	def preventDuplicateWords(self) -> None: ...

	def removeHtmlTagsAll(self) -> None: ...


class GlossaryExtendedType(GlossaryType, typing.Protocol):
	def progressInit(
		self,
		*args,
	) -> None: ...

	def progress(self, pos: int, total: int, unit: str = "entries") -> None: ...

	def progressEnd(self) -> None: ...

	@property
	def progressbar(self) -> bool: ...

	@progressbar.setter
	def progressbar(self, enabled: bool) -> None: ...



# -*- coding: utf-8 -*-

import html
from collections.abc import Iterator
from typing import TYPE_CHECKING

if TYPE_CHECKING:
	import sqlite3

	from pyglossary.glossary_types import EntryType, GlossaryType
	from pyglossary.option import Option

from pyglossary.core import log

__all__ = [
	"Reader",
	"description",
	"enable",
	"extensionCreate",
	"extensions",
	"format",
	"kind",
	"lname",
	"optionsProp",
	"singleFile",
	"website",
	"wiki",
]

enable = True
lname = "dict_cc_split"
format = "Dictcc_split"
description = "Dict.cc (SQLite3) - Split"
extensions = ()
extensionCreate = ".db"
singleFile = True
kind = "binary"
wiki = "https://en.wikipedia.org/wiki/Dict.cc"
website = (
	"https://play.google.com/store/apps/details?id=cc.dict.dictcc",
	"dict.cc dictionary - Google Play",
)
optionsProp: "dict[str, Option]" = {}


class Reader:
	def __init__(self, glos: "GlossaryType") -> None:
		self._glos = glos
		self._clear()

	def _clear(self) -> None:
		self._filename = ""
		self._con: "sqlite3.Connection | None" = None
		self._cur: "sqlite3.Cursor | None" = None

	def open(self, filename: str) -> None:
		from sqlite3 import connect

		self._filename = filename
		self._con = connect(filename)
		self._cur = self._con.cursor()
		self._glos.setDefaultDefiFormat("m")

	def __len__(self) -> int:
		if self._cur is None:
			raise ValueError("cur is None")
		self._cur.execute("select count(*) * 2 from main_ft")
		return self._cur.fetchone()[0]

	def iterRows(
		self,
		column1: str,
		column2: str,
	) -> "Iterator[tuple[str, str, str]]":
		if self._cur is None:
			raise ValueError("cur is None")
		self._cur.execute(
			f"select {column1}, {column2}, entry_type from main_ft"
			f" order by {column1}",
		)
		for row in self._cur.fetchall():
			term1 = row[0]
			term2 = row[1]
			try:
				term1 = html.unescape(term1)
			except Exception as e:
				log.error(f"html.unescape({term1!r}) -> {e}")
			try:
				term2 = html.unescape(term2)
			except Exception as e:
				log.error(f"html.unescape({term2!r}) -> {e}")
			yield term1, term2, row[2]

	def _iterOneDirection(
		self,
		column1: str,
		column2: str,
	) -> "Iterator[EntryType]":
		for word, defi, entry_type in self.iterRows(column1, column2):
			if entry_type:
				word = f"{word} {{{entry_type}}}"  # noqa: PLW2901
			yield self._glos.newEntry(word, defi, defiFormat="m")

	def __iter__(self) -> "Iterator[EntryType]":
		yield from self._iterOneDirection("term1", "term2")
		yield from self._iterOneDirection("term2", "term1")

	def close(self) -> None:
		if self._cur:
			self._cur.close()
		if self._con:
			self._con.close()
		self._clear()



import json
import logging
from os.path import join

from pyglossary.core import rootDir

log = logging.getLogger("pyglossary")


class Lang:
	def __init__(
		self,
		codes: "list[str]",
		names: "list[str]",
		titleTag: str = "b",
		rtl: int = 0,
	) -> None:
		self._codes = codes
		self._names = names
		self._titleTag = titleTag
		self._rtl = rtl

	def __repr__(self) -> str:
		return (
			"Lang("
			f"codes={self._codes!r}, "
			f"names={self._names!r}, "
			f"titleTag={self._titleTag!r}"
			")"
		)

	def __str__(self) -> str:
		return f"Lang({self._codes + self._names})"

	@property
	def codes(self) -> "list[str]":
		return self._codes

	@property
	def names(self) -> "list[str]":
		return self._names

	@property
	def name(self) -> str:
		return self._names[0]

	@property
	def code(self) -> str:
		return self._codes[0]

	@property
	def titleTag(self) -> str:
		return self._titleTag

	@property
	def rtl(self) -> int:
		return self._rtl


class LangDict(dict):
	def _addLang(self, lang: "Lang") -> None:
		for key in lang.codes:
			if key in self:
				log.error(f"duplicate language code: {key}")
			self[key] = lang
		for name in lang.names:
			if name in self:
				log.error(f"duplicate language name: {name}")
			self[name.lower()] = lang

	def load(self) -> None:
		from time import perf_counter as now

		if len(self) > 0:
			return
		t0 = now()
		filename = join(rootDir, "pyglossary", "langs", "langs.json")
		with open(filename, encoding="utf-8") as _file:
			data = json.load(_file)
			for row in data:
				self._addLang(
					Lang(
						codes=row["codes"],
						names=[row["name"]] + row["alt_names"],
						titleTag=row["title_tag"],
						rtl=row.get("rtl", 0),
					),
				)

		log.debug(
			f"LangDict: loaded, {len(self)} keys, "
			f"took {(now() - t0) * 1000:.1f} ms",
		)

	def __getitem__(self, key: str) -> "Lang | None":
		self.load()
		return self.get(key.lower(), None)


langDict = LangDict()



from pyglossary.core import log
from pyglossary.option import EncodingOption, Option, StrOption
from pyglossary.text_reader import TextGlossaryReader

__all__ = [
	"Reader",
	"description",
	"enable",
	"extensionCreate",
	"extensions",
	"format",
	"kind",
	"lname",
	"optionsProp",
	"singleFile",
	"website",
	"wiki",
]

enable = True
lname = "dictunformat"
format = "Dictunformat"
description = "dictunformat output file"
extensions = (".dictunformat",)
extensionCreate = ".dictunformat"
singleFile = True
kind = "text"
wiki = "https://directory.fsf.org/wiki/Dictd"
website = (
	"https://github.com/cheusov/dictd/blob/master/dictunformat.1.in",
	"dictd/dictunformat.1.in - @cheusov/dictd",
)
optionsProp: "dict[str, Option]" = {
	"encoding": EncodingOption(),
	"headword_separator": StrOption(
		comment="separator for headword and alternates",
	),
}


def unescapeDefi(defi: str) -> str:
	return defi


class Reader(TextGlossaryReader):
	_headword_separator = ";   "
	# https://github.com/cheusov/dictd/blob/master/dictfmt/dictunformat.in#L14

	@classmethod
	def isInfoWord(cls, word: str) -> bool:
		return word.startswith("00-database-")

	@classmethod
	def fixInfoWord(cls, word: str) -> str:
		return word

	def setInfo(self, word: str, defi: str) -> None:
		if word == "00-database-short":
			self._glos.setInfo("name", defi)
			return

		if word != "00-database-info":
			return

		glos = self._glos

		lastKey = ""
		for line in defi.split("\n"):
			if not line.startswith("##:"):
				if lastKey:
					glos.setInfo(word, f"{glos.getInfo(lastKey)}\n{line}")
				continue

			parts = line[3:].split(":")
			if len(parts) < 2:
				log.error(f"unexpected line: {line}")
			key = lastKey = parts[0]
			value = ":".join(parts[1:])
			glos.setInfo(key, value)

	def nextBlock(self) -> "tuple[str | list[str], str, None] | None":
		if not self._file:
			raise StopIteration
		word = ""
		defiLines: "list[str]" = []

		while True:
			line = self.readline()
			if not line:
				break
			line = line.rstrip("\n\r")
			if not line:
				continue

			if not line.strip("_"):
				if not word:
					continue
				if not defiLines:
					log.warning(f"no definition/value for {word!r}")
				defi = unescapeDefi("\n".join(defiLines))
				words = word.split(self._headword_separator)
				return words, defi, None

			if not word:
				word = line
				continue

			if line == word:
				continue
			if line.lower() == word:
				word = line
				continue

			defiLines.append(line)

		if word:
			defi = unescapeDefi("\n".join(defiLines))
			if word.startswith("00-database-") and defi == "unknown":
				log.info(f"ignoring {word} -> {defi}")
				return None
			words = word.split(self._headword_separator)
			return words, defi, None

		raise StopIteration



# -*- coding: utf-8 -*-
#
# Copyright © 2022 Saeed Rasooli <saeed.gnu@gmail.com> (ilius)
# This file is part of PyGlossary project, https://github.com/ilius/pyglossary
#
# This program is a free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program. Or on Debian systems, from /usr/share/common-licenses/GPL
# If not, see <http://www.gnu.org/licenses/gpl.txt>.

import logging
from collections import OrderedDict as odict
from typing import TYPE_CHECKING

if TYPE_CHECKING:
	import pathlib
	from typing import Any

	from .flags import StrWithDesc

from . import core
from .flags import (
	DEFAULT_NO,
	YesNoAlwaysNever,
	flagsByName,
)
from .option import Option, optionFromDict

__all__ = ["PluginProp"]

log = logging.getLogger("pyglossary")


def optionsPropFromDict(
	optionsPropDict: "dict[str, Any]",
) -> "dict[str, Option]":
	props = {}
	for name, propDict in optionsPropDict.items():
		try:
			prop = optionFromDict(propDict)
		except Exception:
			log.exception(f"{name=}, {propDict=}\n")
			continue
		props[name] = prop
	return props


def sortOnWriteFromStr(sortOnWriteStr: "str | None") -> "StrWithDesc":
	if sortOnWriteStr is None:
		return DEFAULT_NO
	return flagsByName[sortOnWriteStr]


class PluginCheckError(Exception):
	pass


class PluginProp:  # noqa: PLR0904
	__slots__ = [
		"_Reader",
		"_ReaderLoaded",
		"_Writer",
		"_WriterLoaded",
		"_canRead",
		"_canWrite",
		"_description",
		"_enable",
		"_extensionCreate",
		"_extensions",
		"_lname",
		"_mod",
		"_moduleName",
		"_modulePath",
		"_name",
		"_optionsProp",
		"_readCompressions",
		"_readDepends",
		"_readOptions",
		"_singleFile",
		"_sortKeyName",
		"_sortOnWrite",
		"_writeDepends",
		"_writeOptions",
	]

	def __init__(self) -> None:
		self._mod: "Any"
		self._Reader: "Any"
		self._ReaderLoaded: bool
		self._Writer: "Any"
		self._WriterLoaded: bool

		self._moduleName: str
		self._modulePath: str
		self._enable: bool
		self._lname: str
		self._name: str
		self._description: str
		self._extensions: "list[str]"
		self._extensionCreate: str
		self._singleFile: bool
		self._optionsProp: "dict[str, Option]"
		self._sortOnWrite: YesNoAlwaysNever
		self._sortKeyName: "str | None"
		self._canRead: bool
		self._canWrite: bool
		self._readOptions: "dict[str, Any]"
		self._writeOptions: "dict[str, Any]"
		self._readCompressions: "list[str]"
		self._readDepends: "dict[str, str]"
		self._writeDepends: "dict[str, str]"

	@classmethod
	def fromDict(
		cls: "type",
		attrs: "dict[str, Any]",
		modulePath: str,
	) -> None:
		self = cls()
		self._mod = None
		self._Reader = None
		self._ReaderLoaded = False
		self._Writer = None
		self._WriterLoaded = False

		self._moduleName = attrs["module"]
		self._modulePath = modulePath
		self._enable = attrs.get("enable", True)
		self._lname = attrs["lname"]
		self._name = attrs["name"]
		self._description = attrs["description"]
		self._extensions = attrs["extensions"]
		self._extensionCreate = attrs.get("extensionCreate", "")
		self._singleFile = attrs["singleFile"]
		self._optionsProp = optionsPropFromDict(attrs["optionsProp"])
		self._sortOnWrite = sortOnWriteFromStr(attrs.get("sortOnWrite"))
		self._sortKeyName = attrs.get("sortKeyName")
		self._canRead = attrs["canRead"]
		self._canWrite = attrs["canWrite"]
		self._readOptions = attrs.get("readOptions", {})
		self._writeOptions = attrs.get("writeOptions", {})
		self._readCompressions = attrs.get("readCompressions", [])
		self._readDepends = attrs.get("readDepends", {})
		self._writeDepends = attrs.get("writeDepends", {})

		return self

	@classmethod
	def fromModule(cls: "type", mod: "Any") -> "PluginProp":
		self = cls()
		self._mod = mod
		self._Reader = None
		self._ReaderLoaded = False
		self._Writer = None
		self._WriterLoaded = False

		self._moduleName = mod.__name__
		self._modulePath = mod.__file__
		if self._modulePath.endswith("__init__.py"):
			self._modulePath = self._modulePath[: -len("/__init__.py")]
		elif self._modulePath.endswith(".py"):
			self._modulePath = self._modulePath[:-3]

		self._enable = getattr(mod, "enable", True)
		self._lname = mod.lname
		self._name = mod.format
		self._description = mod.description
		self._extensions = list(mod.extensions)
		self._extensionCreate = getattr(mod, "extensionCreate", "")
		self._singleFile = getattr(mod, "singleFile", False)
		self._optionsProp = getattr(mod, "optionsProp", {})
		self._sortOnWrite = getattr(mod, "sortOnWrite", DEFAULT_NO)
		self._sortKeyName = getattr(mod, "sortKeyName", None)
		self._canRead = hasattr(mod, "Reader")
		self._canWrite = hasattr(mod, "Writer")
		self._readOptions = None
		self._writeOptions = None
		self._readCompressions = None
		self._readDepends = None
		self._writeDepends = None

		if core.isDebug():
			self.checkModule(mod)

		return self

	@property
	def enable(self) -> bool:
		return self._enable

	@property
	def module(self) -> "Any":
		if self._mod is not None:
			return self._mod
		moduleName = self._moduleName
		log.debug(f"importing {moduleName} in DictPluginProp")
		try:
			_mod = __import__(
				f"pyglossary.plugins.{moduleName}",
				fromlist=moduleName,
			)
		except ModuleNotFoundError as e:
			log.warning(
				f"Module {e.name!r} not found in {self._modulePath}"
				f", skipping plugin {moduleName!r}",
			)
			return None
		except Exception:
			log.exception(f"Error while importing plugin {moduleName}")
			return None

		# self._mod = _mod
		if core.isDebug():
			self.checkModule(_mod)

		return _mod

	@property
	def lname(self) -> str:
		return self._lname

	@property
	def name(self) -> str:
		return self._name

	@property
	def description(self) -> str:
		return self._description

	@property
	def extensions(self) -> "list[str]":
		return self._extensions

	@property
	def ext(self) -> str:
		extensions = self.extensions
		if extensions:
			return extensions[0]
		return ""

	@property
	def extensionCreate(self) -> str:
		return self._extensionCreate

	@property
	def singleFile(self) -> bool:
		return self._singleFile

	@property
	def optionsProp(self) -> "dict[str, Option]":
		return self._optionsProp

	@property
	def sortOnWrite(self) -> YesNoAlwaysNever:
		return self._sortOnWrite

	@property
	def sortKeyName(self) -> "str | None":
		return self._sortKeyName

	@property
	def path(self) -> "pathlib.Path":
		from pathlib import Path

		return Path(self._modulePath)

	@property
	def readerClass(self) -> "Any | None":
		if self._ReaderLoaded:
			return self._Reader
		cls = getattr(self.module, "Reader", None)
		self._Reader = cls
		self._ReaderLoaded = True
		if cls is not None and core.isDebug():
			self.checkReaderClass()
		return cls

	@property
	def writerClass(self) -> "Any | None":
		if self._WriterLoaded:
			return self._Writer
		cls = getattr(self.module, "Writer", None)
		self._Writer = cls
		self._WriterLoaded = True
		if cls is not None and core.isDebug():
			self.checkWriterClass()
		return cls

	@property
	def canRead(self) -> bool:
		return self._canRead

	@property
	def canWrite(self) -> bool:
		return self._canWrite

	@staticmethod
	def _getOptionAttrNamesFromClass(rwclass: "type") -> "list[str]":
		nameList = []

		for cls in (*rwclass.__bases__, rwclass):
			for _name in cls.__dict__:
				if not _name.startswith("_") or _name.startswith("__"):
					# and _name not in ("_open",)
					continue
				nameList.append(_name)

		# rwclass.__dict__ does not include attributes of parent/base class
		# and dir(rwclass) is sorted by attribute name alphabetically
		# using rwclass.__bases__ solves the problem

		return nameList

	def _getOptionsFromClass(self, rwclass: "type") -> "dict[str, Any]":
		optionsProp = self.optionsProp
		options = odict()
		if rwclass is None:
			return options

		for attrName in self._getOptionAttrNamesFromClass(rwclass):
			name = attrName[1:]
			default = getattr(rwclass, attrName)
			if name not in optionsProp:
				if not callable(default):
					log.warning(f"format={self.name}, {attrName=}, {type(default)=}")
				continue
			prop = optionsProp[name]
			if prop.disabled:
				core.trace(
					log,
					f"skipping disabled option {name} in {self.name} plugin",
				)
				continue
			if not prop.validate(default):
				log.warning(
					"invalid default value for option: "
					f"{name} = {default!r} in plugin {self.name}",
				)
			options[name] = default

		return options

	def getReadOptions(self) -> "dict[str, Any]":
		if self._readOptions is None:
			self._readOptions = self._getOptionsFromClass(self.readerClass)
		return self._readOptions

	def getWriteOptions(self) -> "dict[str, Any]":
		if self._writeOptions is None:
			self._writeOptions = self._getOptionsFromClass(self.writerClass)
		return self._writeOptions

	@property
	def readCompressions(self) -> "list[str]":
		if self._readCompressions is None:
			self._readCompressions = getattr(self.readerClass, "compressions", [])
		return self._readCompressions

	@property
	def readDepends(self) -> "dict[str, str]":
		if self._readDepends is None:
			self._readDepends = getattr(self.readerClass, "depends", {})
		return self._readDepends

	@property
	def writeDepends(self) -> "dict[str, str]":
		if self._writeDepends is None:
			self._writeDepends = getattr(self.writerClass, "depends", {})
		return self._writeDepends

	def checkModule(self, module) -> None:
		name = self.name

		if hasattr(module, "write"):
			log.error(
				f"plugin {name!r} has write function, must migrate to Writer class",
			)

		extensions = module.extensions
		if not isinstance(extensions, tuple):
			msg = f"{name} plugin: extensions must be tuple"
			if isinstance(extensions, list):
				extensions = tuple(extensions)
				log.error(msg)
			else:
				raise TypeError(msg)

		if not isinstance(self.readDepends, dict):
			log.error(
				f"invalid depends={self.readDepends} in {self.name!r}.Reader class",
			)

		if not isinstance(self.writeDepends, dict):
			log.error(
				f"invalid depends={self.writeDepends}"
				f" in {self.name!r}.Reader class",
			)

		for name, opt in self.optionsProp.items():
			if name.lower() != name:
				suggestName = "".join(
					"_" + x.lower() if x.isupper() else x for x in name
				)
				log.debug(
					f"{self.name}: please rename option {name} to {suggestName}",
				)
			if not opt.comment:
				log.debug(
					f"{self.name}: please add comment for option {name}",
				)

	valid__all__ = [
		"enable",
		"lname",
		"format",
		"description",
		"extensions",
		"extensionCreate",
		"singleFile",
		"kind",
		"wiki",
		"website",
		"optionsProp",
		"Reader",
		"Writer",
	]

	# only run this on CI to do extra validation
	def checkModuleMore(self, module) -> None:
		name = self.name
		if not hasattr(module, "__all__"):
			raise PluginCheckError(f"Please add __all__ to plugin {name!r}")
		_all = module.__all__
		for attr in _all:
			if not hasattr(module, attr):
				raise PluginCheckError(
					f"Undefined name {attr!r} in __all__ in plugin {name!r}"
					f": {module.__file__}",
				)
			if attr not in self.valid__all__:
				raise PluginCheckError(
					f"Unnecessary name {attr!r} in __all__ in plugin {name!r}"
					f": {module.__file__}",
				)

	def checkReaderClass(self) -> bool:
		cls = self._Reader
		for attr in (
			"__init__",
			"open",
			"close",
			"__len__",
			"__iter__",
		):
			if not hasattr(cls, attr):
				log.error(
					f"Invalid Reader class in {self.name!r} plugin"
					f", no {attr!r} method",
				)
				self._Reader = None
				return False

		return True

	def checkWriterClass(self) -> bool:
		cls = self._Writer
		for attr in (
			"__init__",
			"open",
			"write",
			"finish",
		):
			if not hasattr(cls, attr):
				log.error(
					f"Invalid Writer class in {self.name!r} plugin"
					f", no {attr!r} method",
				)
				self._Writer = None
				return False

		return True

	# def _getReadExtraOptions(self) -> "list[str]":  # noqa: F811
	# 	cls = self.readerClass
	# 	if cls is None:
	# 		return []
	# 	return self.__class__.getExtraOptionsFromFunc(cls.open, self.name)

	# def _getWriteExtraOptions(self) -> "list[str]":  # noqa: F811
	# 	cls = self.writerClass
	# 	if cls is None:
	# 		return []
	# 	return self.__class__.getExtraOptionsFromFunc(cls.write, self.name)

	# @classmethod
	# def getExtraOptionsFromFunc(
	# 	cls: "type",
	# 	func: "Callable",
	# 	format: str,
	# ) -> "list[str]":
	# 	import inspect

	# 	extraOptNames = []
	# 	for name, param in inspect.signature(func).parameters.items():
	# 		if name == "self":
	# 			continue
	# 		if str(param.default) != "<class 'inspect._empty'>":
	# 			extraOptNames.append(name)
	# 			continue
	# 		if name not in {"filename", "dirname"}:
	# 			extraOptNames.append(name)
	# 	if extraOptNames:
	# 		log.warning(f"{format}: {extraOptNames = }")
	# 	return extraOptNames



import re
from typing import TYPE_CHECKING, cast

from pyglossary.core import log

from .pinyin import convert
from .summarize import summarize

if TYPE_CHECKING:
	from collections.abc import Sequence

	from pyglossary.lxml_types import T_htmlfile


line_reg = re.compile(r"^([^ ]+) ([^ ]+) \[([^\]]+)\] /(.+)/$")

COLORS = {
	"": "black",
	"1": "red",
	"2": "orange",
	"3": "green",
	"4": "blue",
	"5": "black",
}


def parse_line(line: str) -> "tuple[str, str, str, list[str]] | None":
	line = line.strip()
	match = line_reg.match(line)
	if match is None:
		return None
	trad, simp, pinyin, eng = match.groups()
	pinyin = pinyin.replace("u:", "v")
	return trad, simp, pinyin, eng.split("/")


def make_entry(
	trad: str,
	simp: str,
	pinyin: str,
	eng: list[str],
	traditional_title: bool,
) -> "tuple[list[str], str]":
	eng_names = list(map(summarize, eng))
	names = [
		trad if traditional_title else simp,
		simp if traditional_title else trad,
		pinyin,
	] + eng_names
	article = render_article(trad, simp, pinyin, eng, traditional_title)
	return names, article


def colorize(
	hf: "T_htmlfile",
	syllables: "Sequence[str]",
	tones: "Sequence[str]",
) -> None:
	if len(syllables) != len(tones):
		log.warning(f"unmatched tones: {syllables=}, {tones=}")
		with hf.element("div", style="display: inline-block"):
			for syllable in syllables:
				with hf.element("font", color=""):
					hf.write(syllable)
		return

	with hf.element("div", style="display: inline-block"):
		for syllable, tone in zip(syllables, tones, strict=False):
			with hf.element("font", color=COLORS[tone]):
				hf.write(syllable)


def render_article(
	trad: str,
	simp: str,
	pinyin: str,
	eng: list[str],
	traditional_title: bool,
) -> str:
	from io import BytesIO

	from lxml import etree as ET

	# pinyin_tones = [convert(syl) for syl in pinyin.split()]
	pinyin_list = []
	tones = []
	for syllable in pinyin.split():
		nice_syllable, tone = convert(syllable)
		pinyin_list.append(nice_syllable)
		tones.append(tone)

	f = BytesIO()
	with ET.htmlfile(f, encoding="utf-8") as hf:  # noqa: PLR1702
		with hf.element("div", style="border: 1px solid; padding: 5px"):
			with hf.element("div"):
				with hf.element("big"):
					colorize(
						cast("T_htmlfile", hf),
						trad if traditional_title else simp,
						tones,
					)
				if trad != simp:
					hf.write("\xa0/\xa0")  # "\xa0" --> "&#160;" == "&nbsp;"
					colorize(
						cast("T_htmlfile", hf),
						simp if traditional_title else trad,
						tones,
					)
				hf.write(ET.Element("br"))
				with hf.element("big"):
					colorize(
						cast("T_htmlfile", hf),
						pinyin_list,
						tones,
					)

			with hf.element("div"):
				with hf.element("ul"):
					for defn in eng:
						with hf.element("li"):
							hf.write(defn)

	return f.getvalue().decode("utf-8")



# -*- coding: utf-8 -*-
# dependency.py
#
# Copyright © 2019-2019 Saeed Rasooli <saeed.gnu@gmail.com> (ilius)
# This file is part of PyGlossary project, https://github.com/ilius/pyglossary
#
# This program is a free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program. Or on Debian systems, from /usr/share/common-licenses/GPL
# If not, see <http://www.gnu.org/licenses/gpl.txt>.

# reqs = subprocess.check_output([sys.executable, '-m', 'pip', 'freeze'])
# ^ this takes about 3 seconds
# installed_packages = set(r.decode().split('==')[0] for r in reqs.split())

__all__ = ["checkDepends"]


def checkDepends(depends: "dict[str, str]") -> "list[str]":
	"""Return the list of non-installed dependencies."""
	if not depends:
		return []
	not_installed = []
	for moduleName, pkgName in depends.items():
		try:
			__import__(moduleName)
		except ModuleNotFoundError:  # noqa: PERF203
			not_installed.append(pkgName)
	return not_installed



# -*- coding: utf-8 -*-

import html
from collections.abc import Iterator
from typing import TYPE_CHECKING

if TYPE_CHECKING:
	import sqlite3

	from pyglossary.glossary_types import EntryType, GlossaryType
	from pyglossary.option import Option

__all__ = [
	"Reader",
	"description",
	"enable",
	"extensionCreate",
	"extensions",
	"format",
	"kind",
	"lname",
	"optionsProp",
	"singleFile",
	"website",
	"wiki",
]

enable = True
lname = "digitalnk"
format = "DigitalNK"
description = "DigitalNK (SQLite3, N-Korean)"
extensions = ()
extensionCreate = ".db"
singleFile = True
kind = "binary"
wiki = ""
website = (
	"https://github.com/digitalprk/dicrs",
	"@digitalprk/dicrs",
)
optionsProp: "dict[str, Option]" = {}


class Reader:
	def __init__(self, glos: "GlossaryType") -> None:
		self._glos = glos
		self._clear()

	def _clear(self) -> None:
		self._filename = ""
		self._con: "sqlite3.Connection | None" = None
		self._cur: "sqlite3.Cursor | None" = None

	def open(self, filename: str) -> None:
		from sqlite3 import connect

		self._filename = filename
		self._con = connect(filename)
		self._cur = self._con.cursor()
		self._glos.setDefaultDefiFormat("m")

	def __len__(self) -> int:
		if self._cur is None:
			raise ValueError("cur is None")
		self._cur.execute("select count(*) from dictionary")
		return self._cur.fetchone()[0]

	def __iter__(self) -> "Iterator[EntryType]":
		if self._cur is None:
			raise ValueError("cur is None")
		self._cur.execute(
			"select word, definition from dictionary order by word",
		)
		# iteration over self._cur stops after one entry
		# and self._cur.fetchone() returns None
		# no idea why!
		# https://github.com/ilius/pyglossary/issues/282
		# for row in self._cur:
		for row in self._cur.fetchall():
			word = html.unescape(row[0])
			definition = row[1]
			yield self._glos.newEntry(word, definition, defiFormat="m")

	def close(self) -> None:
		if self._cur:
			self._cur.close()
		if self._con:
			self._con.close()
		self._clear()



# -*- coding: utf-8 -*-
#
# Output to Apple Dictionary xml sources for Dictionary Development Kit.
#
# Copyright © 2016-2023 Saeed Rasooli <saeed.gnu@gmail.com> (ilius)
# Copyright © 2016 ivan tkachenko <me@ratijas.tk>
# Copyright © 2012-2015 Xiaoqiang Wang <xiaoqiangwang AT gmail DOT com>
#
# This program is a free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3 of the License.
#
# You can get a copy of GNU General Public License along this program
# But you can always get it from http://www.gnu.org/licenses/gpl.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.

import os
import pkgutil
import shutil
import sys
from os.path import basename, isdir, join
from typing import TYPE_CHECKING, Any

from pyglossary.core import log, pip
from pyglossary.glossary_types import EntryType, GlossaryType
from pyglossary.option import (
	BoolOption,
	DictOption,
	Option,
	StrOption,
)
from pyglossary.text_utils import toStr

from ._content import prepare_content
from ._dict import (
	_normalize,
	id_generator,
	indexes_generator,
	quote_string,
)

if TYPE_CHECKING:
	import io
	from collections.abc import Generator

__all__ = [
	"Writer",
	"description",
	"enable",
	"extensionCreate",
	"extensions",
	"format",
	"kind",
	"lname",
	"optionsProp",
	"singleFile",
	"website",
	"wiki",
]

sys.setrecursionlimit(10000)

enable = True
lname = "appledict"
format = "AppleDict"
description = "AppleDict Source"
extensions = (".apple",)
extensionCreate = ".apple/"
singleFile = False
kind = "directory"
wiki = ""
website = (
	"https://support.apple.com/en-gu/guide/dictionary/welcome/mac",
	"Dictionary User Guide for Mac",
)
# FIXME: rename indexes arg/option to indexes_lang?
optionsProp: "dict[str, Option]" = {
	"clean_html": BoolOption(comment="use BeautifulSoup parser"),
	"css": StrOption(
		comment="custom .css file path",
	),
	"xsl": StrOption(
		comment="custom XSL transformations file path",
	),
	"default_prefs": DictOption(
		comment="default prefs in python dict format",
		# example: {"key": "value", "version": "1"}
	),
	"prefs_html": StrOption(
		comment="preferences XHTML file path",
	),
	"front_back_matter": StrOption(
		comment="XML file path with top-level tag",
	),
	"jing": BoolOption(comment="run Jing check on generated XML"),
	"indexes": StrOption(
		customValue=False,
		values=["", "ru", "zh"],
		comment="Additional indexes to dictionary entries",
	),
}

extraDocs = [
	(
		"Also see:",
		"See [doc/apple.md](./doc/apple.md) for additional AppleDict instructions.",
	),
]


BeautifulSoup = None


def loadBeautifulSoup() -> None:
	global BeautifulSoup
	try:
		import bs4 as BeautifulSoup
	except ImportError:
		try:
			import BeautifulSoup  # type: ignore
		except ImportError:
			return
	_version: str = BeautifulSoup.__version__  # type: ignore
	if int(_version.split(".")[0]) < 4:
		raise ImportError(
			"BeautifulSoup is too old, required at least version 4, "
			f"{_version!r} found.\n"
			f"Please run `{pip} install lxml beautifulsoup4 html5lib`",
		)


def abspath_or_None(path: "str | None") -> "str | None":
	if not path:
		return None
	return os.path.abspath(os.path.expanduser(path))


def write_header(
	toFile: "io.TextIOBase",
	front_back_matter: "str | None",
) -> None:
	# write header
	toFile.write(
		'<?xml version="1.0" encoding="UTF-8"?>\n'
		'<d:dictionary xmlns="http://www.w3.org/1999/xhtml" '
		'xmlns:d="http://www.apple.com/DTDs/DictionaryService-1.0.rng">\n',
	)

	if front_back_matter:
		with open(
			front_back_matter,
			encoding="utf-8",
		) as _file:
			toFile.write(_file.read())


def format_default_prefs(default_prefs: "dict[str, Any] | None") -> str:
	"""
	:type default_prefs: dict or None

	as by 14th of Jan 2016, it is highly recommended that prefs should contain
	{"version": "1"}, otherwise Dictionary.app does not keep user changes
	between restarts.
	"""
	if not default_prefs:
		return ""
	if not isinstance(default_prefs, dict):
		raise TypeError(f"default_prefs not a dictionary: {default_prefs!r}")
	if str(default_prefs.get("version", None)) != "1":
		log.error(
			"default prefs does not contain {'version': '1'}.  prefs "
			"will not be persistent between Dictionary.app restarts.",
		)
	return "\n".join(
		f"\t\t<key>{key}</key>\n\t\t<string>{value}</string>"
		for key, value in sorted(default_prefs.items())
	).strip()


def write_css(fname: str, css_file: str) -> None:
	with open(fname, mode="wb") as toFile:
		if css_file:
			with open(css_file, mode="rb") as fromFile:
				toFile.write(fromFile.read())
		else:
			data = pkgutil.get_data(
				__name__,
				"templates/Dictionary.css",
			)
			if data is None:
				raise RuntimeError("failed to load templates/Dictionary.css")
			toFile.write(data)


"""
write glossary to Apple dictionary .xml and supporting files.

:param dirname: directory path, must not have extension

:param clean_html: pass True to use BeautifulSoup parser.

:param css: path to custom .css file

:param xsl: path to custom XSL transformations file.

:param default_prefs: Default prefs in python dictionary literal format,
i.e. {"key1": "value1", "key2": "value2", ...}.  All keys and values
must be quoted strings; not allowed characters (e.g. single/double
quotes,equal sign "=", semicolon) must be escaped as hex code
according to python string literal rules.

:param prefs_html: path to XHTML file with user interface for
dictionary's preferences. refer to Apple's documentation for details.

:param front_back_matter: path to XML file with top-level tag
<d:entry id="front_back_matter" d:title="Your Front/Back Matter Title">
	your front/back matter entry content
</d:entry>

:param jing: pass True to run Jing check on generated XML.

# FIXME: rename to indexes_lang?
:param indexes: Dictionary.app is dummy and by default it don't know
how to perform flexible search.  we can help it by manually providing
additional indexes to dictionary entries.
"""


class Writer:
	depends = {
		"lxml": "lxml",
		"bs4": "beautifulsoup4",
		"html5lib": "html5lib",
	}

	_clean_html: bool = True
	_css: str = ""
	_xsl: str = ""
	_default_prefs: "dict | None" = None
	_prefs_html: str = ""
	_front_back_matter: str = ""
	_jing: bool = False
	_indexes: str = ""  # FIXME: rename to indexes_lang?

	def __init__(self, glos: GlossaryType) -> None:
		self._glos = glos
		self._dirname = ""

	def finish(self) -> None:
		self._dirname = ""

	def open(self, dirname: str) -> None:
		self._dirname = dirname
		if not isdir(dirname):
			os.mkdir(dirname)

	def write(self) -> "Generator[None, EntryType, None]":  # noqa: PLR0912
		global BeautifulSoup
		from pyglossary.xdxf.transform import XdxfTransformer

		glos = self._glos
		clean_html = self._clean_html
		css: "str | None" = self._css
		xsl: "str | None" = self._xsl
		default_prefs = self._default_prefs
		prefs_html: "str | None" = self._prefs_html
		front_back_matter: "str | None" = self._front_back_matter
		jing = self._jing
		indexes = self._indexes

		xdxf_to_html = XdxfTransformer(encoding="utf-8")

		if clean_html:
			if BeautifulSoup is None:
				loadBeautifulSoup()
			if BeautifulSoup is None:
				log.warning(
					"clean_html option passed but BeautifulSoup not found. "
					"to fix this run "
					f"`{pip} install lxml beautifulsoup4 html5lib`",
				)
		else:
			BeautifulSoup = None

		dirname = self._dirname
		fileNameBase = basename(dirname).replace(".", "_")
		filePathBase = join(dirname, fileNameBase)
		# before chdir (outside indir block)
		css = abspath_or_None(css)
		xsl = abspath_or_None(xsl)
		prefs_html = abspath_or_None(prefs_html)
		front_back_matter = abspath_or_None(front_back_matter)

		generate_id = id_generator()
		generate_indexes = indexes_generator(indexes)

		myResDir = join(dirname, "OtherResources")
		if not isdir(myResDir):
			os.mkdir(myResDir)

		with open(filePathBase + ".xml", mode="w", encoding="utf-8") as toFile:
			write_header(toFile, front_back_matter)
			while True:
				entry = yield
				if entry is None:
					break
				if entry.isData():
					entry.save(myResDir)
					continue

				words = entry.l_word
				word, alts = words[0], words[1:]
				defi = entry.defi

				long_title = _normalize.title_long(
					_normalize.title(word, BeautifulSoup),
				)
				if not long_title:
					continue

				_id = next(generate_id)
				quoted_title = quote_string(long_title, BeautifulSoup)

				content_title: "str | None" = long_title
				if entry.defiFormat == "x":
					defi = xdxf_to_html.transformByInnerString(defi)
					content_title = None
				content = prepare_content(content_title, defi, BeautifulSoup)

				toFile.write(
					f'<d:entry id="{_id}" d:title={quoted_title}>\n'
					+ generate_indexes(long_title, alts, content, BeautifulSoup)
					+ content
					+ "\n</d:entry>\n",
				)

			toFile.write("</d:dictionary>\n")

		if xsl:
			shutil.copy(xsl, myResDir)

		if prefs_html:
			shutil.copy(prefs_html, myResDir)

		write_css(filePathBase + ".css", css)

		with open(join(dirname, "Makefile"), mode="w", encoding="utf-8") as toFile:
			toFile.write(
				toStr(
					pkgutil.get_data(
						__name__,
						"templates/Makefile",
					),
				).format(dict_name=fileNameBase),
			)

		_copyright = glos.getInfo("copyright")
		if BeautifulSoup:
			# strip html tags
			_copyright = str(
				BeautifulSoup.BeautifulSoup(
					_copyright,
					features="lxml",
				).text,
			)

		# if DCSDictionaryXSL provided but DCSDictionaryDefaultPrefs <dict/> not
		# present in Info.plist, Dictionary.app will crash.
		with open(filePathBase + ".plist", mode="w", encoding="utf-8") as toFile:
			frontMatterReferenceID = (
				"<key>DCSDictionaryFrontMatterReferenceID</key>\n"
				"\t<string>front_back_matter</string>"
				if front_back_matter
				else ""
			)
			bundle_id = glos.getInfo("CFBundleIdentifier")
			if not bundle_id:
				bundle_id = fileNameBase.replace(" ", "")
			toFile.write(
				toStr(
					pkgutil.get_data(
						__name__,
						"templates/Info.plist",
					),
				).format(
					# identifier must be unique
					CFBundleIdentifier=bundle_id,
					CFBundleDisplayName=glos.getInfo("name"),
					CFBundleName=fileNameBase,
					DCSDictionaryCopyright=_copyright,
					DCSDictionaryManufacturerName=glos.author,
					DCSDictionaryXSL=basename(xsl) if xsl else "",
					DCSDictionaryDefaultPrefs=format_default_prefs(default_prefs),
					DCSDictionaryPrefsHTML=basename(prefs_html) if prefs_html else "",
					DCSDictionaryFrontMatterReferenceID=frontMatterReferenceID,
				),
			)

		if jing:
			from .jing import run as jing_run

			jing_run(filePathBase + ".xml")



# -*- coding: utf-8 -*-

import os
import re
from collections.abc import Generator, Iterator
from os.path import isdir, splitext

from pyglossary.core import log
from pyglossary.flags import DEFAULT_NO
from pyglossary.glossary_types import EntryType, GlossaryType
from pyglossary.option import BoolOption, Option
from pyglossary.plugin_lib.dictdlib import DictDB

__all__ = [
	"Reader",
	"Writer",
	"description",
	"enable",
	"extensionCreate",
	"extensions",
	"format",
	"kind",
	"lname",
	"optionsProp",
	"singleFile",
	"website",
	"wiki",
]

enable = True
lname = "dict_org"
format = "DictOrg"
description = "DICT.org file format (.index)"
extensions = (".index",)
extensionCreate = ""
singleFile = False
optionsProp: "dict[str, Option]" = {
	"dictzip": BoolOption(comment="Compress .dict file to .dict.dz"),
	"install": BoolOption(comment="Install dictionary to /usr/share/dictd/"),
}
sortOnWrite = DEFAULT_NO
kind = "directory"
wiki = "https://en.wikipedia.org/wiki/DICT#DICT_file_format"
website = (
	"http://dict.org/bin/Dict",
	"The DICT Development Group",
)


def installToDictd(filename: str, dictzip: bool) -> None:
	"""Filename is without extension (neither .index or .dict or .dict.dz)."""
	import shutil
	import subprocess

	targetDir = "/usr/share/dictd/"
	if filename.startswith(targetDir):
		return

	if not isdir(targetDir):
		log.warning(f"Directory {targetDir!r} does not exist, skipping install")
		return

	log.info(f"Installing {filename!r} to DICTD server directory: {targetDir}")

	if dictzip and os.path.isfile(filename + ".dict.dz"):
		dictExt = ".dict.dz"
	elif os.path.isfile(filename + ".dict"):
		dictExt = ".dict"
	else:
		log.error(f"No .dict file, could not install dictd file {filename!r}")
		return

	if not filename.startswith(targetDir):
		shutil.copy(filename + ".index", targetDir)
		shutil.copy(filename + dictExt, targetDir)

	# update /var/lib/dictd/db.list
	if subprocess.call(["/usr/sbin/dictdconfig", "-w"]) != 0:
		log.error(
			"failed to update /var/lib/dictd/db.list file"
			", try manually running: sudo /usr/sbin/dictdconfig -w",
		)

	log.info("don't forget to restart dictd server")


class Reader:
	def __init__(self, glos: GlossaryType) -> None:
		self._glos = glos
		self._filename = ""
		self._dictdb: "DictDB | None" = None

		# regular expression patterns used to prettify definition text
		self._re_newline_in_braces = re.compile(
			r"\{(?P<left>.*?)\n(?P<right>.*?)?\}",
		)
		self._re_words_in_braces = re.compile(
			r"\{(?P<word>.+?)\}",
		)

	def open(self, filename: str) -> None:
		filename = filename.removesuffix(".index")
		self._filename = filename
		self._dictdb = DictDB(filename, "read", 1)

	def close(self) -> None:
		if self._dictdb is not None:
			self._dictdb.close()
			# self._dictdb.finish()
			self._dictdb = None

	def prettifyDefinitionText(self, defi: str) -> str:
		# Handle words in {}
		# First, we remove any \n in {} pairs
		defi = self._re_newline_in_braces.sub(r"{\g<left>\g<right>}", defi)

		# Then, replace any {words} into <a href="bword://words">words</a>,
		# so it can be rendered as link correctly
		defi = self._re_words_in_braces.sub(
			r'<a href="bword://\g<word>">\g<word></a>',
			defi,
		)

		# Use <br /> so it can be rendered as newline correctly
		return defi.replace("\n", "<br />")

	def __len__(self) -> int:
		if self._dictdb is None:
			return 0
		return len(self._dictdb)

	def __iter__(self) -> "Iterator[EntryType]":
		if self._dictdb is None:
			raise RuntimeError("iterating over a reader while it's not open")
		dictdb = self._dictdb
		for word in dictdb.getDefList():
			b_defi = b"\n\n<hr>\n\n".join(dictdb.getDef(word))
			try:
				defi = b_defi.decode("utf_8", "ignore")
				defi = self.prettifyDefinitionText(defi)
			except Exception as e:
				log.error(f"{b_defi = }")
				raise e
			yield self._glos.newEntry(word, defi)


class Writer:
	_dictzip: bool = False
	_install: bool = True

	def __init__(self, glos: GlossaryType) -> None:
		self._glos = glos
		self._filename = ""
		self._dictdb: "DictDB | None" = None

	def finish(self) -> None:
		from pyglossary.os_utils import runDictzip

		if self._dictdb is None:
			raise RuntimeError("self._dictdb is None")

		self._dictdb.finish(dosort=True)
		if self._dictzip:
			runDictzip(f"{self._filename}.dict")
		if self._install:
			installToDictd(
				self._filename,
				self._dictzip,
			)
		self._filename = ""

	def open(self, filename: str) -> None:
		filename_nox, ext = splitext(filename)
		if ext.lower() == ".index":
			filename = filename_nox
		self._dictdb = DictDB(filename, "write", 1)
		self._filename = filename

	def write(self) -> "Generator[None, EntryType, None]":
		dictdb = self._dictdb
		if dictdb is None:
			raise RuntimeError("self._dictdb is None")
		while True:
			entry = yield
			if entry is None:
				break
			if entry.isData():
				# does dictd support resources? and how? FIXME
				continue
			dictdb.addEntry(entry.defi, entry.l_word)
